import sys
from typing import Optional

import pytest
from pytest_test_utils import TmpDir

from scmrepo.git import Git, Stash


def test_git_stash_workspace(tmp_dir: TmpDir, scm: Git):
    tmp_dir.gen({"file": "0"})
    scm.add_commit("file", message="init")
    tmp_dir.gen("file", "1")

    with scm.stash_workspace():
        assert not scm.is_dirty()
        assert (tmp_dir / "file").read_text() == "0"
    assert scm.is_dirty()
    assert (tmp_dir / "file").read_text() == "1"


def test_git_stash_workspace_reinstate_index(tmp_dir: TmpDir, scm: Git):
    tmp_dir.gen({"modified": "init", "deleted": "deleted"})
    scm.add_commit(["modified", "deleted"], "init")

    tmp_dir.gen({"newfile": "nefile"})
    scm.add("newfile")
    tmp_dir.gen({"modified": "modified"})
    scm.add("modified")
    (tmp_dir / "deleted").unlink()
    scm.add("deleted")

    prev_status = scm.status()
    with scm.stash_workspace(reinstate_index=True):
        pass
    assert scm.status() == prev_status


@pytest.mark.parametrize(
    "ref, include_untracked",
    [
        (None, True),
        (None, False),
        ("refs/foo/stash", True),
        ("refs/foo/stash", False),
    ],
)
def test_git_stash_push(
    tmp_dir: TmpDir, scm: Git, ref: Optional[str], include_untracked: bool
):
    tmp_dir.gen({"file": "0"})
    scm.add_commit("file", message="init")
    tmp_dir.gen({"file": "1", "untracked": "0"})

    stash = Stash(scm, ref=ref)
    rev = stash.push(include_untracked=include_untracked)
    assert rev == scm.get_ref(stash.ref)
    assert (tmp_dir / "file").read_text() == "0"
    assert include_untracked != (tmp_dir / "untracked").exists()
    assert len(stash) == 1

    stash.apply(rev)
    assert (tmp_dir / "file").read_text() == "1"
    assert (tmp_dir / "untracked").read_text() == "0"

    parts = list(stash.ref.split("/"))
    assert (tmp_dir / ".git").joinpath(*parts).exists()
    assert (tmp_dir / ".git" / "logs").joinpath(*parts).exists()


@pytest.mark.parametrize("ref", [None, "refs/foo/stash"])
def test_git_stash_drop(tmp_dir: TmpDir, scm: Git, ref: Optional[str]):
    tmp_dir.gen({"file": "0"})
    scm.add_commit("file", message="init")
    tmp_dir.gen("file", "1")

    stash = Stash(scm, ref=ref)
    stash.push()

    tmp_dir.gen("file", "2")
    expected = stash.push()

    stash.drop(1)
    assert expected == scm.get_ref(stash.ref)
    assert len(stash) == 1


reason = """libgit2 stash_save() is flaky on linux when run inside pytest
    https://github.com/treeverse/dvc/pull/5286#issuecomment-792574294"""


@pytest.mark.xfail(
    sys.platform in ("linux", "win32"),
    raises=AssertionError,
    strict=False,
    reason=reason,
)
@pytest.mark.parametrize("ref", [None, "refs/foo/stash"])
def test_git_stash_pop(tmp_dir: TmpDir, scm: Git, ref: Optional[str]):
    tmp_dir.gen({"file": "0"})
    scm.add_commit("file", message="init")
    tmp_dir.gen("file", "1")

    stash = Stash(scm, ref=ref)
    first = stash.push()

    tmp_dir.gen("file", "2")
    second = stash.push()

    assert second == stash.pop()
    assert len(stash) == 1
    assert first == scm.get_ref(stash.ref)
    assert (tmp_dir / "file").read_text() == "2"


@pytest.mark.parametrize("ref", [None, "refs/foo/stash"])
def test_git_stash_clear(tmp_dir: TmpDir, scm: Git, ref: Optional[str]):
    tmp_dir.gen({"file": "0"})
    scm.add_commit("file", message="init")
    tmp_dir.gen("file", "1")

    stash = Stash(scm, ref=ref)
    stash.push()

    tmp_dir.gen("file", "2")
    stash.push()

    stash.clear()
    assert len(stash) == 0

    parts = list(stash.ref.split("/"))
    assert not (tmp_dir / ".git").joinpath(*parts).exists()

    reflog_file = (tmp_dir / ".git" / "logs").joinpath(*parts)
    # NOTE: some backends will completely remove reflog file on clear, some
    # will only truncate it, either case means an empty stash
    assert not reflog_file.exists() or not reflog_file.cat()


@pytest.mark.skip_git_backend("dulwich")
def test_git_stash_apply_index(
    tmp_dir: TmpDir,
    scm: Git,
    git: Git,
):
    tmp_dir.gen("file", "0")
    scm.add_commit("file", message="init")
    tmp_dir.gen("file", "1")
    scm.add("file")
    scm.stash.push()
    rev = scm.resolve_rev(r"stash@{0}")

    stash = Stash(git)
    stash.apply(rev, reinstate_index=True)

    assert (tmp_dir / "file").read_text() == "1"
    staged, unstaged, untracked = scm.status()
    assert dict(staged) == {"modify": ["file"]}
    assert not dict(unstaged)
    assert not dict(untracked)


def test_git_stash_push_clean_workspace(
    tmp_dir: TmpDir,
    scm: Git,
    git: Git,
):
    tmp_dir.gen("file", "0")
    scm.add_commit("file", message="init")
    assert git._stash_push("refs/stash") == (  # pylint: disable=protected-access
        None,
        False,
    )



import os
import socket
import threading
from io import StringIO
from typing import Any
from unittest.mock import AsyncMock

import asyncssh
import paramiko
import pytest
from paramiko.server import InteractiveQuery
from pytest_mock import MockerFixture
from pytest_test_utils.waiters import wait_until

from scmrepo.exceptions import AuthError
from scmrepo.git.backend.dulwich.asyncssh_vendor import AsyncSSHVendor

# pylint: disable=redefined-outer-name


USER = "testuser"
PASSWORD = "test"
CLIENT_KEY = """-----BEGIN RSA PRIVATE KEY-----
MIIEpAIBAAKCAQEAxvREKSElPOm/0z/nPO+j5rk2tjdgGcGc7We1QZ6TRXYLu7nN
GeEFIL4p8N1i6dmB+Eydt7xqCU79MWD6Yy4prFe1+/K1wCDUxIbFMxqQcX5zjJzd
i8j8PbcaUlVhP/OkjtkSxrXaGDO1BzfdV4iEBtTV/2l3zmLKJlt3jnOHLczP24CB
DTQKp3rKshbRefzot9Y+wnaK692RsYgsyo9YEP0GyWKG9topCHk13r46J6vGLeuj
ryUKqmbLJkzbJbIcEqwTDo5iHaCVqaMr5Hrb8BdMucSseqZQJsXSd+9tdRcIblUQ
38kZjmFMm4SFbruJcpZCNM2wNSZPIRX+3eiwNwIDAQABAoIBAHSacOBSJsr+jIi5
KUOTh9IPtzswVUiDKwARCjB9Sf8p4lKR4N1L/n9kNJyQhApeikgGT2GCMftmqgoo
tlculQoHFgemBlOmak0MV8NNzF5YKEy/GzF0CDH7gJfEpoyetVFrdA+2QS5yD6U9
XqKQxiBi2VEqdScmyyeT8AwzNYTnPeH/DOEcnbdRjqiy/CD79F49CQ1lX1Fuqm0K
I7BivBH1xo/rVnUP4F+IzocDqoga+Pjdj0LTXIgJlHQDSbhsQqWujWQDDuKb+MAw
sNK4Zf8ErV3j1PyA7f/M5LLq6zgstkW4qikDHo4SpZX8kFOO8tjqb7kujj7XqeaB
CxqrOTECgYEA73uWkrohcmDJ4KqbuL3tbExSCOUiaIV+sT1eGPNi7GCmXD4eW5Z4
75v2IHymW83lORSu/DrQ6sKr1nkuRpqr2iBzRmQpl/H+wahIhBXlnJ25uUjDsuPO
1Pq2LcmyD+jTxVnmbSe/q7O09gZQw3I6H4+BMHmpbf8tC97lqimzpJ0CgYEA1K0W
ZL70Xtn9quyHvbtae/BW07NZnxvUg4UaVIAL9Zu34JyplJzyzbIjrmlDbv6aRogH
/KtuG9tfbf55K/jjqNORiuRtzt1hUN1ye4dyW7tHx2/7lXdlqtyK40rQl8P0kqf8
zaS6BqjnobgSdSpg32rWoL/pcBHPdJCJEgQ8zeMCgYEA0/PK8TOhNIzrP1dgGSKn
hkkJ9etuB5nW5mEM7gJDFDf6JPupfJ/xiwe6z0fjKK9S57EhqgUYMB55XYnE5iIw
ZQ6BV9SAZ4V7VsRs4dJLdNC3tn/rDGHJBgCaym2PlbsX6rvFT+h1IC8dwv0V79Ui
Ehq9WTzkMoE8yhvNokvkPZUCgYEAgBAFxv5xGdh79ftdtXLmhnDvZ6S8l6Fjcxqo
Ay/jg66Tp43OU226iv/0mmZKM8Dd1xC8dnon4GBVc19jSYYiWBulrRPlx0Xo/o+K
CzZBN1lrXH1i6dqufpc0jq8TMf/N+q1q/c1uMupsKCY1/xVYpc+ok71b7J7c49zQ
nOeuUW8CgYA9Infooy65FTgbzca0c9kbCUBmcAPQ2ItH3JcPKWPQTDuV62HcT00o
fZdIV47Nez1W5Clk191RMy8TXuqI54kocciUWpThc6j44hz49oUueb8U4bLcEHzA
WxtWBWHwxfSmqgTXilEA3ALJp0kNolLnEttnhENwJpZHlqtes0ZA4w==
-----END RSA PRIVATE KEY-----"""


class Server(paramiko.ServerInterface):
    """http://docs.paramiko.org/en/2.4/api/server.html."""

    def __init__(self, commands, *args, **kwargs) -> None:
        super().__init__()
        self.commands = commands
        self.allowed_auths = kwargs.get("allowed_auths", "publickey,password")

    def check_channel_exec_request(self, channel, command):
        self.commands.append(command)
        return True

    def check_auth_interactive(self, username: str, submethods: str):
        return InteractiveQuery(
            "Password", "Enter the password", f"Password for user {USER}:"
        )

    def check_auth_interactive_response(self, responses):
        if responses[0] == PASSWORD:
            return paramiko.AUTH_SUCCESSFUL  # type: ignore[attr-defined]
        return paramiko.AUTH_FAILED  # type: ignore[attr-defined]

    def check_auth_password(self, username, password):
        if username == USER and password == PASSWORD:
            return paramiko.AUTH_SUCCESSFUL  # type: ignore[attr-defined]
        return paramiko.AUTH_FAILED  # type: ignore[attr-defined]

    def check_auth_publickey(self, username, key):
        pubkey = paramiko.RSAKey.from_private_key(StringIO(CLIENT_KEY))
        if username == USER and key == pubkey:
            return paramiko.AUTH_SUCCESSFUL  # type: ignore[attr-defined]
        return paramiko.AUTH_FAILED  # type: ignore[attr-defined]

    def check_channel_request(self, kind, chanid):
        if kind == "session":
            return paramiko.OPEN_SUCCEEDED  # type: ignore[attr-defined]
        return paramiko.OPEN_FAILED_ADMINISTRATIVELY_PROHIBITED  # type: ignore[attr-defined]

    def get_allowed_auths(self, username):
        return self.allowed_auths


@pytest.fixture
def ssh_conn(request: pytest.FixtureRequest) -> dict[str, Any]:
    server = Server([], **getattr(request, "param", {}))

    socket.setdefaulttimeout(10)
    request.addfinalizer(lambda: socket.setdefaulttimeout(None))

    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.bind(("127.0.0.1", 0))
    sock.listen(5)
    request.addfinalizer(sock.close)  # noqa: PT021
    port = sock.getsockname()[1]

    conn_info: dict[str, Any] = {"port": port, "server": server}

    def _run_server():
        try:
            conn, _ = sock.accept()
        except OSError:
            return False
        server.transport = transport = paramiko.Transport(conn)  # type: ignore[attr-defined]
        request.addfinalizer(transport.close)
        host_key = paramiko.RSAKey.from_private_key(StringIO(CLIENT_KEY))
        transport.add_server_key(host_key)
        transport.start_server(server=server)

    thread = threading.Thread(target=_run_server)
    thread.start()
    return conn_info


@pytest.fixture
def ssh_port(ssh_conn: dict[str, Any]) -> int:
    return ssh_conn["port"]


@pytest.fixture
def server(ssh_conn: dict[str, Any]) -> Server:
    return ssh_conn["server"]


def test_run_command_password(server: Server, ssh_port: int):
    vendor = AsyncSSHVendor()
    vendor.run_command(
        "127.0.0.1",
        "test_run_command_password",
        username=USER,
        port=ssh_port,
        password=PASSWORD,
    )

    assert b"test_run_command_password" in server.commands


@pytest.mark.parametrize("ssh_conn", [{"allowed_auths": "publickey"}], indirect=True)
def test_run_command_no_password(ssh_port: int):
    vendor = AsyncSSHVendor()
    with pytest.raises(AuthError):
        vendor.run_command(
            "127.0.0.1",
            "test_run_command_password",
            username=USER,
            port=ssh_port,
            password=None,
        )


@pytest.mark.parametrize(
    "ssh_conn",
    [{"allowed_auths": "password"}, {"allowed_auths": "keyboard-interactive"}],
    indirect=True,
    ids=["password", "interactive"],
)
def test_should_prompt_for_password_when_no_password_passed(
    mocker: MockerFixture, server: Server, ssh_port: int
):
    mocked_getpass = mocker.patch("getpass.getpass", return_value=PASSWORD)
    vendor = AsyncSSHVendor()
    vendor.run_command(
        "127.0.0.1",
        "test_run_command_password",
        username=USER,
        port=ssh_port,
        password=None,
    )
    assert server.commands == [b"test_run_command_password"]
    mocked_getpass.asssert_called_once()


def test_run_command_with_privkey(server: Server, ssh_port: int):
    key = asyncssh.import_private_key(CLIENT_KEY)

    vendor = AsyncSSHVendor()
    vendor.run_command(
        "127.0.0.1",
        "test_run_command_with_privkey",
        username=USER,
        port=ssh_port,
        key_filename=key,
    )

    assert b"test_run_command_with_privkey" in server.commands


@pytest.mark.parametrize("use_len", [True, False])
def test_run_command_data_transfer(server: Server, ssh_port: int, use_len: bool):
    vendor = AsyncSSHVendor()
    con = vendor.run_command(
        "127.0.0.1",
        "test_run_command_data_transfer",
        username=USER,
        port=ssh_port,
        password=PASSWORD,
    )

    assert b"test_run_command_data_transfer" in server.commands

    channel = server.transport.accept(5)  # type: ignore[attr-defined]
    channel.send(b"stdout\n")
    channel.send_stderr(b"stderr\n")
    channel.close()

    assert wait_until(con.can_read, timeout=1, pause=0.1)
    assert con.read(n=7 if use_len else None) == b"stdout\n"
    assert con.read_stderr(n=7 if use_len else None) == b"stderr\n"


def test_run_command_partial_transfer(ssh_port: int, mocker: MockerFixture):
    vendor = AsyncSSHVendor()
    con = vendor.run_command(
        "127.0.0.1",
        "test_run_command_data_transfer",
        username=USER,
        port=ssh_port,
        password=PASSWORD,
    )

    mock_stdout = mocker.patch.object(
        con.proc.stdout,
        "read",
        side_effect=[b"s", b"tdout", b"\n"],
        new_callable=AsyncMock,
    )
    assert con.read(n=7) == b"stdout\n"
    assert mock_stdout.call_count == 3

    mock_stderr = mocker.patch.object(
        con.stderr.stderr,
        "read",
        side_effect=[b"s", b"tderr", b"\n"],
        new_callable=AsyncMock,
    )
    assert con.read_stderr(n=7) == b"stderr\n"
    assert mock_stderr.call_count == 3


@pytest.mark.skipif(os.name != "nt", reason="Windows only")
def test_git_bash_ssh_vendor(mocker):
    from dulwich.client import SubprocessSSHVendor

    from scmrepo.git.backend.dulwich import _get_ssh_vendor

    mocker.patch.dict(os.environ, {"MSYSTEM": "MINGW64"})
    assert isinstance(_get_ssh_vendor(), SubprocessSSHVendor)

    del os.environ["MSYSTEM"]
    assert isinstance(_get_ssh_vendor(), AsyncSSHVendor)


def test_unsupported_config_ssh_vendor():
    from dulwich.client import SubprocessSSHVendor

    from scmrepo.git.backend.dulwich import _get_ssh_vendor

    config = os.path.expanduser(os.path.join("~", ".ssh", "config"))
    os.makedirs(os.path.dirname(config), exist_ok=True)

    with open(config, "wb") as fobj:
        fobj.write(
            b"""
Host *
    IdentityFile ~/.ssh/id_rsa
"""
        )
    assert isinstance(_get_ssh_vendor(), AsyncSSHVendor)

    with open(config, "wb") as fobj:
        fobj.write(
            b"""
Host *
    UseKeychain yes
"""
        )
    assert isinstance(_get_ssh_vendor(), SubprocessSSHVendor)



import fnmatch
import io
import os
import re
from collections.abc import Iterable, Iterator
from typing import TYPE_CHECKING, Callable, Optional

from scmrepo.exceptions import InvalidRemote, SCMError

from .pointer import HEADERS, Pointer

if TYPE_CHECKING:
    from scmrepo.git import Git
    from scmrepo.git.config import Config
    from scmrepo.progress import GitProgressEvent


def fetch(
    scm: "Git",
    revs: Optional[list[str]] = None,
    remote: Optional[str] = None,
    include: Optional[list[str]] = None,
    exclude: Optional[list[str]] = None,
    progress: Optional[Callable[["GitProgressEvent"], None]] = None,
):
    # NOTE: This currently does not support fetching objects from the worktree
    if not revs:
        revs = ["HEAD"]
    objects: set[Pointer] = set()
    for rev in revs:
        objects.update(
            pointer
            for pointer in _collect_objects(scm, rev, include, exclude)
            if not scm.lfs_storage.exists(pointer)
        )
    if not objects:
        return
    try:
        url = get_fetch_url(scm, remote=remote)
    except InvalidRemote:
        if remote:
            # treat remote as a raw Git remote
            url = remote
        else:
            raise
    scm.lfs_storage.fetch(url, objects, progress=progress)


def get_fetch_url(scm: "Git", remote: Optional[str] = None):  # noqa: C901
    """Return LFS fetch URL for the specified repository."""
    git_config = scm.get_config()

    # check lfs.url (can be set in git config and .lfsconfig)
    try:
        return git_config.get(("lfs",), "url")
    except KeyError:
        pass
    try:
        lfs_config: Optional[Config] = scm.get_config(
            os.path.join(scm.root_dir, ".lfsconfig")
        )
    except FileNotFoundError:
        lfs_config = None
    if lfs_config:
        try:
            return lfs_config.get(("lfs",), "url")
        except KeyError:
            pass

    # use:
    #   current tracking-branch remote
    #   or remote.lfsdefault  (can only be set in git config)
    #   or "origin"
    # in that order
    if not remote:
        try:
            remote = scm.active_branch_remote()
        except SCMError:
            pass
    if not remote:
        try:
            remote = git_config.get(("remote",), "lfsdefault")
        except KeyError:
            remote = "origin"

    # check remote.*.lfsurl (can be set in git config and .lfsconfig)
    assert remote is not None
    try:
        return git_config.get(("remote", remote), "lfsurl")
    except KeyError:
        pass
    if lfs_config:
        try:
            return lfs_config.get(("remote", remote), "lfsurl")
        except KeyError:
            pass

    # return default Git fetch URL for this remote
    return scm.get_remote_url(remote)


_ROOT_PATH_PREFIX_REGEX = re.compile(r"^(?P<prefix>[^*?\[]*(?:/|$))")


def _collect_objects(
    scm: "Git",
    rev: str,
    include: Optional[list[str]],
    exclude: Optional[list[str]],
) -> Iterator[Pointer]:
    fs = scm.get_fs(rev)
    # Optimize path filtering if the `include` list contains exactly one path.
    # First, determine the root directory wherein to initiate the file search.
    # If the `include` path is a Unix filename pattern, determine the static
    # path prefix and set it as the root directory. Second, if the path and the
    # root are identical or the Unix filename pattern matches *any* (i.e., `**`)
    # file under the root directory, unset `include` to avoid unnecessary
    # filtering work.
    if (
        include
        and len(include) == 1
        and (result := _ROOT_PATH_PREFIX_REGEX.match(path := include[0]))
    ):
        root = result.group("prefix")
        if path in {root, f"{root.rstrip('/')}/**"}:
            include = []
    else:
        root = "/"

    for path in _filter_paths(fs.find(root), include, exclude):
        check_path = path.lstrip("/")
        if scm.check_attr(check_path, "filter", source=rev) == "lfs":
            try:
                with fs.open(path, "rb", raw=True) as fobj:
                    with io.BufferedReader(fobj) as reader:
                        data = reader.peek(100)
                        if any(data.startswith(header) for header in HEADERS):
                            yield Pointer.load(reader)
            except (ValueError, OSError):
                pass


def _filter_paths(
    paths: Iterable[str], include: Optional[list[str]], exclude: Optional[list[str]]
) -> Iterator[str]:
    filtered = set()
    if include:
        for pattern in include:
            filtered.update(fnmatch.filter(paths, pattern))
    else:
        filtered.update(paths)
    if exclude:
        for pattern in exclude:
            filtered.difference_update(fnmatch.filter(paths, pattern))
    yield from filtered


if __name__ == "__main__":
    # Minimal `git lfs fetch` CLI implementation
    import argparse

    from scmrepo.git import Git

    parser = argparse.ArgumentParser(
        description=(
            "Download Git LFS objects at the given refs from the specified remote."
        ),
    )
    parser.add_argument(
        "remote",
        nargs="?",
        default="origin",
        help="Remote to fetch from. Defaults to 'origin'.",
    )
    parser.add_argument(
        "refs",
        nargs="*",
        default=["HEAD"],
        help="Refs or commits to fetch. Defaults to 'HEAD'.",
    )
    args = parser.parse_args()
    with Git(".") as scm_:  # pylint: disable=E0601
        fetch(scm_, revs=args.refs, remote=args.remote)



# pylint: disable=unused-argument
import os

import pygit2
import pytest
from pytest_mock import MockerFixture
from pytest_test_utils import TmpDir

from scmrepo.exceptions import SCMError
from scmrepo.git import Git
from scmrepo.git.backend.pygit2 import Pygit2Backend


@pytest.mark.parametrize("use_sha", [True, False])
def test_pygit_resolve_refish(tmp_dir: TmpDir, scm: Git, use_sha: str):
    backend = Pygit2Backend(tmp_dir)
    tmp_dir.gen("foo", "foo")
    scm.add_commit("foo", message="foo")
    head = scm.get_rev()
    tag = "my_tag"
    scm.tag(tag, annotated=True, message="create annotated tag")

    if use_sha:
        # refish will be annotated tag SHA (not commit SHA)
        _ref = backend.repo.references.get(f"refs/tags/{tag}")
        assert _ref
        refish = str(_ref.target)
    else:
        refish = tag

    assert refish != head
    commit, ref = backend._resolve_refish(refish)  # pylint: disable=protected-access
    assert isinstance(commit, pygit2.Commit)
    assert str(commit.id) == head
    if not use_sha:
        assert ref.name == f"refs/tags/{tag}"


@pytest.mark.parametrize("skip_conflicts", [True, False])
def test_pygit_stash_apply_conflicts(
    tmp_dir: TmpDir, scm: Git, skip_conflicts: bool, mocker: MockerFixture
):
    from pygit2 import GIT_CHECKOUT_ALLOW_CONFLICTS  # type: ignore[attr-defined]

    tmp_dir.gen("foo", "foo")
    scm.add_commit("foo", message="foo")
    tmp_dir.gen("foo", "bar")
    scm.stash.push()
    rev = scm.resolve_rev(r"stash@{0}")

    backend = Pygit2Backend(tmp_dir)
    mock = mocker.patch.object(backend.repo, "stash_apply")
    backend._stash_apply(  # pylint: disable=protected-access
        rev, skip_conflicts=skip_conflicts
    )
    expected_strategy = (
        backend._get_checkout_strategy()  # pylint: disable=protected-access
    )
    if skip_conflicts:
        expected_strategy |= GIT_CHECKOUT_ALLOW_CONFLICTS
    mock.assert_called_once_with(
        0,
        strategy=expected_strategy,
        reinstate_index=False,
    )


@pytest.mark.parametrize(
    "url",
    [
        "git@github.com:treeverse/scmrepo.git",
        "github.com:treeverse/scmrepo.git",
        "user@github.com:treeverse/scmrepo.git",
        "ssh://login@server.com:12345/repository.git",
    ],
)
def test_pygit_ssh_error(tmp_dir: TmpDir, scm: Git, url):
    backend = Pygit2Backend(tmp_dir)
    with pytest.raises(NotImplementedError):
        with backend._get_remote(url):  # pylint: disable=protected-access
            pass


@pytest.mark.parametrize("name", ["committer", "author"])
def test_pygit_use_env_vars_for_signature(
    tmp_dir: TmpDir, mocker: MockerFixture, name: str
):
    from pygit2 import Signature

    mocker.patch(
        "scmrepo.git.Pygit2Backend.default_signature",
        new=mocker.PropertyMock(side_effect=SCMError),
    )
    git = Git.init(tmp_dir)
    with pytest.raises(SCMError):
        _ = git.pygit2.default_signature  # pylint: disable=W0104

    # Make sure that the environment variables are not set to not interfere with
    # with the check below
    for var in [f"GIT_{name.upper()}_NAME", f"GIT_{name.upper()}_EMAIL"]:
        assert os.environ.get(var, None) is None

    # Basic expected behavior if vars are not set. Another sanity check
    with pytest.raises(SCMError):
        getattr(git.pygit2, name)

    mocker.patch.dict(os.environ, {f"GIT_{name.upper()}_EMAIL": "olivaw@treeverse.io"})
    with pytest.raises(SCMError):
        getattr(git.pygit2, name)

    mocker.patch.dict(os.environ, {f"GIT_{name.upper()}_NAME": "R. Daneel Olivaw"})
    assert getattr(git.pygit2, name) == Signature(
        email="olivaw@treeverse.io", name="R. Daneel Olivaw"
    )



import fnmatch
import locale
import logging
import os
import re
import stat
from collections.abc import Iterable, Iterator, Mapping
from contextlib import closing
from functools import partial
from io import BytesIO, StringIO
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Optional,
    Union,
)

from dulwich.config import ConfigFile, StackedConfig
from dulwich.walk import ORDER_DATE
from funcy import cached_property, reraise

from scmrepo.exceptions import AuthError, CloneError, InvalidRemote, RevError, SCMError
from scmrepo.git.backend.base import BaseGitBackend, SyncStatus
from scmrepo.git.config import Config
from scmrepo.git.objects import GitObject, GitTag
from scmrepo.progress import GitProgressReporter
from scmrepo.utils import relpath

if TYPE_CHECKING:
    from dulwich.client import SSHVendor
    from dulwich.config import ConfigFile
    from dulwich.repo import Repo

    from scmrepo.git.objects import GitCommit
    from scmrepo.progress import GitProgressEvent


logger = logging.getLogger(__name__)


class DulwichObject(GitObject):
    def __init__(self, repo, name, mode, sha) -> None:
        self.repo = repo
        self._name = name
        self._mode = mode
        self._sha = sha

    def open(  # pylint: disable=unused-argument
        self,
        mode: str = "r",
        encoding: Optional[str] = None,
        raw: bool = True,
        rev: Optional[str] = None,
        **kwargs,
    ):
        if not raw:
            raise NotImplementedError
        if not encoding:
            encoding = locale.getpreferredencoding(False)
        # NOTE: we didn't load the object before as Dulwich will also try to
        # load the contents of it into memory, which will slow down Trie
        # building considerably.
        obj = self.repo[self._sha]
        data = obj.as_raw_string()
        if mode == "rb":
            return BytesIO(data)
        return StringIO(data.decode(encoding))

    @property
    def name(self) -> str:
        return self._name

    @property
    def mode(self) -> int:
        return self._mode

    def scandir(self) -> Iterable["DulwichObject"]:
        tree = self.repo[self._sha]
        for entry in tree.iteritems():
            yield DulwichObject(self.repo, entry.path.decode(), entry.mode, entry.sha)

    @cached_property
    def size(self) -> int:  # pylint: disable=invalid-overridden-method
        try:
            return self.repo[self._sha].raw_length()
        except KeyError:
            return 0

    @property
    def sha(self) -> str:
        return self._sha.decode("ascii")


class DulwichProgressReporter(GitProgressReporter):
    """Dulwich progress reporter.

    Works with both dulwich.porcelain methods which expect an 'errstream'
    stream object and internal dulwich methods which expect a 'progress' stream
    write method.
    """

    def write(self, msg: Union[str, bytes]) -> int:
        self(msg)
        return len(msg)


def _get_ssh_vendor() -> "SSHVendor":
    import shutil
    import sys

    from dulwich.client import SubprocessSSHVendor

    from .asyncssh_vendor import AsyncSSHVendor, get_unsupported_opts

    ssh_command = os.environ.get("GIT_SSH_COMMAND", os.environ.get("GIT_SSH"))
    if ssh_command:
        logger.debug("dulwich: Using environment GIT_SSH_COMMAND '%s'", ssh_command)
        return SubprocessSSHVendor()

    if sys.platform == "win32" and os.environ.get("MSYSTEM") and shutil.which("ssh"):
        # see https://github.com/treeverse/dvc/issues/7702
        logger.debug(
            "dulwich: native win32 Python inside MSYS2/git-bash, using MSYS2 OpenSSH"
        )
        return SubprocessSSHVendor()

    default_config = os.path.expanduser(os.path.join("~", ".ssh", "config"))
    unsupported = list(get_unsupported_opts([default_config]))
    if unsupported and shutil.which("ssh"):
        logger.debug(
            "dulwich: unsupported SSH config option(s) '%s', using system OpenSSH",
            ", ".join(unsupported),
        )
        return SubprocessSSHVendor()
    return AsyncSSHVendor()


class DulwichConfig(Config):
    def __init__(self, config: Union["ConfigFile", "StackedConfig"]) -> None:
        self._config = config

    @property
    def encoding(self) -> str:
        from dulwich.config import ConfigFile

        if isinstance(self._config, ConfigFile):
            return self._config.encoding
        return self._config.backends[0].encoding

    def get(self, section: tuple[str, ...], name: str) -> str:
        """Return the specified setting as a string."""
        return self._config.get(section, name).decode(self.encoding)

    def get_bool(self, section: tuple[str, ...], name: str) -> bool:
        """Return the specified setting as a boolean."""
        value = self._config.get_boolean(section, name)
        if value is None:
            raise ValueError("setting is not a valid boolean")
        return value

    def get_multivar(self, section: tuple[str, ...], name: str) -> Iterator[str]:
        """Iterate over string values in the specified multivar setting."""
        for value in self._config.get_multivar(section, name):
            yield value.decode(self.encoding)


class DulwichBackend(BaseGitBackend):  # pylint:disable=abstract-method
    """Dulwich Git backend."""

    from dulwich import client

    from .client import GitCredentialsHTTPClient

    # monkeypatch dulwich client's default SSH vendor to use asyncssh
    client.get_ssh_vendor = _get_ssh_vendor  # type: ignore[assignment]
    # monkeypatch dulwich client's default HTTPClient to add support for
    # git credential helpers. See https://github.com/jelmer/dulwich/pull/976
    client.HttpGitClient = GitCredentialsHTTPClient  # type: ignore[assignment]

    # Dulwich progress will return messages equivalent to git CLI,
    # our pbars should just display the messages as formatted by dulwich
    BAR_FMT_NOTOTAL = "{desc}{bar:b}|{postfix[info]} [{elapsed}]"

    def __init__(  # pylint:disable=W0231
        self, root_dir=os.curdir, search_parent_directories=True
    ) -> None:
        from dulwich.errors import NotGitRepository
        from dulwich.repo import Repo

        try:
            if search_parent_directories:
                self.repo = Repo.discover(start=root_dir)
            else:
                self.repo = Repo(root_dir)
        except NotGitRepository as exc:
            raise SCMError(f"{root_dir} is not a git repository") from exc

        self._submodules: dict[str, str] = self._find_submodules()
        self._stashes: dict = {}

    def list_submodules(self) -> list[str]:
        raise NotImplementedError

    def _find_submodules(self) -> dict[str, str]:
        """Return dict mapping submodule names to submodule paths.

        Submodule paths will be relative to Git repo root.
        """

        from dulwich.config import ConfigFile, parse_submodules

        submodules: dict[str, str] = {}
        assert self.root_dir
        config_path = os.path.join(self.root_dir, ".gitmodules")
        if os.path.isfile(config_path):
            config = ConfigFile.from_path(config_path)
            for path, _url, section in parse_submodules(config):
                submodules[os.fsdecode(section)] = os.fsdecode(path)
        return submodules

    def close(self):
        self.repo.close()

    @property
    def root_dir(self) -> Optional[str]:
        return self.repo.path

    @classmethod
    def clone(
        cls,
        url: str,
        to_path: Union[str, os.PathLike[str]],
        shallow_branch: Optional[str] = None,
        progress: Optional[Callable[["GitProgressEvent"], None]] = None,
        bare: bool = False,
        mirror: bool = False,
    ):
        from urllib.parse import urlparse

        from dulwich.porcelain import NoneStream
        from dulwich.porcelain import clone as git_clone

        if mirror:
            bare = True
        parsed = urlparse(url)
        try:
            clone_from = partial(
                git_clone,
                url,
                target=to_path,
                errstream=(
                    DulwichProgressReporter(progress) if progress else NoneStream()  # type: ignore[arg-type]
                ),
                bare=bare,
            )
            if shallow_branch:
                # NOTE: dulwich only supports shallow/depth for non-local
                # clones. This differs from CLI git, where depth is used for
                # file:// URLs but not direct local paths
                if parsed.scheme in ("git", "git+ssh", "ssh", "http", "https"):
                    depth = 1
                else:
                    depth = 0
                repo = clone_from(depth=depth, branch=os.fsencode(shallow_branch))
            else:
                repo = clone_from()

            with closing(repo):
                if mirror:
                    cls._set_mirror(repo, progress=progress)
                else:
                    cls._set_default_tracking_branch(repo)
        except Exception as exc:
            raise CloneError(url, os.fsdecode(to_path)) from exc

    @staticmethod
    def _set_default_tracking_branch(repo: "Repo"):
        from dulwich.refs import LOCAL_BRANCH_PREFIX, parse_symref_value

        head_ref = repo.refs.read_ref(b"HEAD")
        if head_ref is None:
            return
        try:
            ref = parse_symref_value(head_ref)
        except ValueError:
            return
        if ref.startswith(LOCAL_BRANCH_PREFIX):
            branch = ref[len(LOCAL_BRANCH_PREFIX) :]
            config = repo.get_config()
            section = ("branch", os.fsencode(branch))
            config.set(section, b"remote", b"origin")
            config.set(section, b"merge", ref)

    @staticmethod
    def _set_mirror(
        repo: "Repo", progress: Optional[Callable[["GitProgressEvent"], None]] = None
    ):
        from dulwich.porcelain import NoneStream, fetch

        config = repo.get_config()
        section = config[(b"remote", b"origin")]
        try:
            del section[b"fetch"]
        except KeyError:
            pass
        section[b"fetch"] = b"+refs/*:refs/*"
        section[b"mirror"] = b"true"
        config.write_to_path()
        fetch(
            repo,
            remote_location=b"origin",
            errstream=(DulwichProgressReporter(progress) if progress else NoneStream()),  # type: ignore[arg-type]
        )

    @staticmethod
    def init(path: str, bare: bool = False) -> None:
        from dulwich.porcelain import init

        init(path, bare=bare)

    @property
    def dir(self) -> str:
        return self.repo.commondir()

    def add(
        self,
        paths: Union[str, Iterable[str]],
        update: bool = False,
        force: bool = False,
    ):
        assert paths or update

        paths = [paths] if isinstance(paths, str) else list(paths)

        worktree = self.repo.get_worktree()

        if update and not paths:
            worktree.stage(list(self.repo.open_index()))
            return

        files: list[bytes] = [
            os.fsencode(fpath) for fpath in self._expand_paths(paths, force=force)
        ]
        if update:
            index = self.repo.open_index()
            if os.name == "nt":
                # NOTE: we need git/unix separator to compare against index
                # paths but repo.stage() expects to be called with OS paths
                worktree.stage(
                    [fname for fname in files if fname.replace(b"\\", b"/") in index]
                )
            else:
                worktree.stage([fname for fname in files if fname in index])
        else:
            worktree.stage(files)

    def _expand_paths(self, paths: list[str], force: bool = False) -> Iterator[str]:
        for path in paths:
            if os.path.isdir(path):
                for root, _, fs in os.walk(path):
                    for fpath in fs:
                        rel = relpath(os.path.join(root, fpath), self.root_dir)
                        if force or not self.ignore_manager.is_ignored(rel):
                            yield rel
            else:
                rel = relpath(path, self.root_dir)
                if force or not self.ignore_manager.is_ignored(rel):
                    yield rel

    def commit(self, msg: str, no_verify: bool = False):
        from dulwich.errors import CommitError
        from dulwich.porcelain import Error, TimezoneFormatError, commit
        from dulwich.repo import InvalidUserIdentity

        with reraise((Error, CommitError), SCMError("Git commit failed")):
            try:
                commit(self.repo, message=msg, no_verify=no_verify, sign=False)
            except InvalidUserIdentity as exc:
                raise SCMError("Git username and email must be configured") from exc
            except TimezoneFormatError as exc:
                raise SCMError("Invalid Git timestamp") from exc

    def checkout(
        self,
        branch: str,
        create_new: Optional[bool] = False,
        force: bool = False,
        **kwargs,
    ):
        raise NotImplementedError

    def fetch(
        self,
        remote: Optional[str] = None,
        force: bool = False,
        unshallow: bool = False,
    ):
        from dulwich.porcelain import Error, fetch
        from dulwich.protocol import DEPTH_INFINITE

        with reraise(Error, SCMError("Git fetch failed")):
            remote_b = os.fsencode(remote) if remote else b"origin"
            fetch(
                self.repo,
                remote_location=remote_b,
                force=force,
                depth=DEPTH_INFINITE if unshallow else None,
            )

    def pull(self, **kwargs):
        raise NotImplementedError

    def push(self):
        raise NotImplementedError

    def branch(self, branch: str):
        from dulwich.porcelain import Error, branch_create

        try:
            branch_create(self.repo, branch)
        except Error as exc:
            raise SCMError(f"Failed to create branch '{branch}'") from exc

    def tag(
        self,
        tag: str,
        target: Optional[str] = None,
        annotated: bool = False,
        message: Optional[str] = None,
    ):
        from dulwich.porcelain import Error, tag_create

        if annotated and not message:
            raise SCMError("message is required for annotated tag")
        with reraise(Error, SCMError("Failed to create tag")):
            tag_create(
                self.repo,
                os.fsencode(tag),
                objectish=target or "HEAD",
                annotated=annotated,
                message=message.encode("utf-8") if message else None,
            )

    def untracked_files(self) -> Iterable[str]:
        _staged, _unstaged, untracked = self.status()
        return untracked

    def is_tracked(self, path: str) -> bool:
        rel = relpath(path, self.root_dir).replace(os.path.sep, "/").encode()
        rel_dir = rel + b"/"
        return any(p == rel or p.startswith(rel_dir) for p in self.repo.open_index())

    def is_dirty(self, untracked_files: bool = False) -> bool:
        kwargs: dict[str, Any] = {} if untracked_files else {"untracked_files": "no"}
        return any(self.status(**kwargs))

    def active_branch(self) -> str:
        raise NotImplementedError

    def active_branch_remote(self) -> str:
        raise NotImplementedError

    def list_branches(self) -> Iterable[str]:
        base = "refs/heads/"
        return sorted(ref[len(base) :] for ref in self.iter_refs(base))

    def list_tags(self) -> Iterable[str]:
        base = "refs/tags/"
        return sorted(ref[len(base) :] for ref in self.iter_refs(base))

    def list_all_commits(self) -> Iterable[str]:
        from dulwich.objects import Tag

        repo = self.repo
        starting_points: list[bytes] = []

        # HEAD
        head_rev = self.get_ref("HEAD")
        if head_rev:
            starting_points.append(head_rev.encode("utf-8"))

        # Branches and remotes
        for ref in repo.refs:
            if ref.startswith((b"refs/heads/", b"refs/remotes/", b"refs/tags/")):
                if isinstance(repo.refs[ref], Tag):
                    ref = self.repo.get_peeled(repo.refs[ref])
                starting_points.append(repo.refs[ref])

        walker = self.repo.get_walker(include=starting_points, order=ORDER_DATE)
        return [e.commit.id.decode() for e in walker]

    def get_tree_obj(self, rev: str, **kwargs) -> DulwichObject:
        from dulwich.objectspec import parse_tree

        tree = parse_tree(self.repo, rev)
        return DulwichObject(self.repo, ".", stat.S_IFDIR, tree.id)

    def get_rev(self) -> str:
        rev = self.get_ref("HEAD")
        if rev:
            return rev
        raise SCMError("Empty git repo")

    def resolve_rev(self, rev: str) -> str:
        raise NotImplementedError

    def resolve_commit(self, rev: str) -> "GitCommit":
        raise NotImplementedError

    def _get_stash(self, ref: str):
        from dulwich.stash import Stash as DulwichStash

        if ref not in self._stashes:
            self._stashes[ref] = DulwichStash(self.repo, ref=os.fsencode(ref))
        return self._stashes[ref]

    @cached_property
    def ignore_manager(self):
        from dulwich.ignore import IgnoreFilterManager

        return IgnoreFilterManager.from_repo(self.repo)

    def is_ignored(self, path: "Union[str, os.PathLike[str]]") -> bool:
        # `is_ignored` returns `false` if excluded in `.gitignore` and
        # `None` if it's not mentioned at all. `True` if it is ignored.
        relative_path = relpath(path, self.root_dir)
        # if checking a directory, a trailing slash must be included
        if str(path)[-1] == os.sep:
            relative_path += os.sep
        return bool(self.ignore_manager.is_ignored(relative_path))

    def set_ref(
        self,
        name: str,
        new_ref: str,
        old_ref: Optional[str] = None,
        message: Optional[str] = None,
        symbolic: Optional[bool] = False,
    ):
        name_b = os.fsencode(name)
        new_ref_b = os.fsencode(new_ref)
        old_ref_b = os.fsencode(old_ref) if old_ref else None
        message_b = message.encode("utf-8") if message else None
        if symbolic:
            return self.repo.refs.set_symbolic_ref(name_b, new_ref_b, message=message_b)
        if not self.repo.refs.set_if_equals(
            name_b, old_ref_b, new_ref_b, message=message_b
        ):
            raise SCMError(f"Failed to set '{name}'")

    def get_ref(self, name, follow: bool = True) -> Optional[str]:
        from dulwich.objects import Tag
        from dulwich.refs import parse_symref_value

        name_b = os.fsencode(name)
        if follow:
            try:
                ref = self.repo.refs[name_b]
            except KeyError:
                ref = None
        else:
            ref = self.repo.refs.read_ref(name_b)
            try:
                if ref:
                    ref = parse_symref_value(ref)
            except ValueError:
                pass
        if ref:
            if ref in self.repo and isinstance(self.repo[ref], Tag):
                ref = self.repo.get_peeled(name_b)
            return os.fsdecode(ref)
        return None

    def remove_ref(self, name: str, old_ref: Optional[str] = None):
        name_b = name.encode("utf-8")
        old_ref_b = old_ref.encode("utf-8") if old_ref else None
        if not self.repo.refs.remove_if_equals(name_b, old_ref_b):
            raise SCMError(f"Failed to remove '{name}'")

    def iter_refs(self, base: Optional[str] = None):
        base_b = os.fsencode(base) if base else None
        for key in self.repo.refs.keys(base=base_b):
            if base:
                if base.endswith("/"):
                    base = base[:-1]
                yield "/".join([base, os.fsdecode(key)])
            else:
                yield os.fsdecode(key)

    def iter_remote_refs(self, url: str, base: Optional[str] = None, **kwargs):
        from dulwich.client import HTTPUnauthorized, get_transport_and_path
        from dulwich.errors import NotGitRepository
        from dulwich.porcelain import get_remote_repo

        try:
            _remote, location = get_remote_repo(self.repo, url)
            _config = kwargs.pop("config", StackedConfig.default())
            client, path = get_transport_and_path(location, config=_config, **kwargs)
        except Exception as exc:
            raise InvalidRemote(url) from exc

        try:
            refs = client.get_refs(path.encode()).refs

            if base:
                base_b = os.fsencode(base)
                yield from (os.fsdecode(ref) for ref in refs if ref.startswith(base_b))
            else:
                yield from (os.fsdecode(ref) for ref in refs)
        except NotGitRepository as exc:
            raise InvalidRemote(url) from exc
        except HTTPUnauthorized as exc:
            raise AuthError(url) from exc

    def get_refs_containing(self, rev: str, pattern: Optional[str] = None):
        raise NotImplementedError

    def push_refspecs(  # noqa: C901
        self,
        url: str,
        refspecs: Union[str, Iterable[str]],
        force: bool = False,
        on_diverged: Optional[Callable[[str, str], bool]] = None,
        progress: Optional[Callable[["GitProgressEvent"], None]] = None,
        **kwargs,
    ) -> Mapping[str, SyncStatus]:
        from dulwich.client import HTTPUnauthorized, get_transport_and_path
        from dulwich.errors import NotGitRepository, SendPackError
        from dulwich.objectspec import parse_reftuples
        from dulwich.porcelain import DivergedBranches, check_diverged, get_remote_repo

        try:
            _remote, location = get_remote_repo(self.repo, url)
            _config = kwargs.pop("config", StackedConfig.default())
            client, path = get_transport_and_path(location, config=_config, **kwargs)
        except Exception as exc:
            raise SCMError(f"'{url}' is not a valid Git remote or URL") from exc

        change_result = {}
        selected_refs: list[tuple[Optional[bytes], Optional[bytes], bool]] = []

        def update_refs(refs):
            from dulwich.objects import ZERO_SHA

            _refspecs = (
                os.fsencode(refspecs)
                if isinstance(refspecs, str)
                else [os.fsencode(refspec) for refspec in refspecs]
            )
            selected_refs.extend(
                parse_reftuples(self.repo.refs, refs, _refspecs, force=force)
            )
            new_refs = {}
            for lh, rh, _ in selected_refs:
                assert rh is not None
                refname = os.fsdecode(rh)
                if rh in refs and lh is not None:
                    if refs[rh] == self.repo.refs[lh]:
                        change_result[refname] = SyncStatus.UP_TO_DATE
                        continue
                    try:
                        check_diverged(self.repo, refs[rh], self.repo.refs[lh])
                    except DivergedBranches:
                        if not force:
                            overwrite = (
                                on_diverged(os.fsdecode(lh), os.fsdecode(refs[rh]))
                                if on_diverged
                                else False
                            )
                            if not overwrite:
                                change_result[refname] = SyncStatus.DIVERGED
                                continue

                if lh is None:
                    value = ZERO_SHA
                else:
                    value = self.repo.refs[lh]

                new_refs[rh] = value
                change_result[refname] = SyncStatus.SUCCESS

            return new_refs

        try:
            result = client.send_pack(
                path.encode(),
                update_refs,
                generate_pack_data=self.repo.generate_pack_data,  # type: ignore[arg-type]
                progress=(DulwichProgressReporter(progress) if progress else None),
            )
        except (NotGitRepository, SendPackError) as exc:
            src = [lh for (lh, _, _) in selected_refs]
            raise SCMError(f"Git failed to push '{src}' to '{url}'") from exc
        except HTTPUnauthorized as exc:
            raise AuthError(url) from exc
        if result.ref_status and any(
            (value is not None) for value in result.ref_status.values()
        ):
            reasons = ", ".join(
                (
                    f"{os.fsdecode(ref)}: {reason}"
                    for ref, reason in result.ref_status.items()
                    if reason is not None
                )
            )
            raise SCMError(f"Git failed to push some refs to '{url}' ({reasons})")
        return change_result

    def fetch_refspecs(
        self,
        url: str,
        refspecs: Union[str, Iterable[str]],
        force: bool = False,
        on_diverged: Optional[Callable[[str, str], bool]] = None,
        progress: Optional[Callable[["GitProgressEvent"], None]] = None,
        **kwargs,
    ) -> Mapping[str, SyncStatus]:
        from dulwich.client import get_transport_and_path
        from dulwich.errors import NotGitRepository
        from dulwich.objectspec import parse_reftuples
        from dulwich.porcelain import DivergedBranches, check_diverged, get_remote_repo
        from dulwich.refs import DictRefsContainer

        fetch_refs: list[tuple[Optional[bytes], Optional[bytes], bool]] = []

        def determine_wants(
            remote_refs: dict[bytes, bytes],
            depth: Optional[int] = None,  # pylint: disable=unused-argument
        ) -> list[bytes]:
            fetch_refs.extend(
                parse_reftuples(
                    DictRefsContainer(remote_refs),
                    self.repo.refs,
                    os.fsencode(refspecs)
                    if isinstance(refspecs, str)
                    else [os.fsencode(refspec) for refspec in refspecs],
                    force=force,
                )
            )
            return [
                remote_refs[lh]
                for (lh, _, _) in fetch_refs
                if lh is not None and remote_refs[lh] not in self.repo.object_store
            ]

        with reraise(Exception, SCMError(f"'{url}' is not a valid Git remote or URL")):
            _remote, location = get_remote_repo(self.repo, url)
            _config = kwargs.pop("config", StackedConfig.default())
            client, path = get_transport_and_path(location, config=_config, **kwargs)

        with reraise(
            (NotGitRepository, KeyError),
            SCMError(f"Git failed to fetch ref from '{url}'"),
        ):
            fetch_result = client.fetch(
                path.encode(),
                self.repo,
                progress=DulwichProgressReporter(progress) if progress else None,
                determine_wants=determine_wants,  # type: ignore[arg-type]
            )

            result = {}

            for lh, rh, _ in fetch_refs:
                assert rh is not None
                refname = os.fsdecode(rh)
                assert lh is not None
                fetch_ref_lh = fetch_result.refs[lh]
                assert fetch_ref_lh is not None
                if rh in self.repo.refs:
                    if self.repo.refs[rh] == fetch_ref_lh:
                        result[refname] = SyncStatus.UP_TO_DATE
                        continue
                    try:
                        check_diverged(
                            self.repo,
                            self.repo.refs[rh],
                            fetch_ref_lh,
                        )
                    except DivergedBranches:
                        if not force:
                            overwrite = (
                                on_diverged(os.fsdecode(rh), os.fsdecode(fetch_ref_lh))
                                if on_diverged
                                else False
                            )
                            if not overwrite:
                                result[refname] = SyncStatus.DIVERGED
                                continue

                self.repo.refs[rh] = fetch_ref_lh
                result[refname] = SyncStatus.SUCCESS
        return result

    def _stash_iter(self, ref: str):
        stash = self._get_stash(ref)
        yield from stash.stashes()

    def _stash_push(
        self,
        ref: str,
        message: Optional[str] = None,
        include_untracked: bool = False,
    ) -> tuple[Optional[str], bool]:
        from dulwich.repo import InvalidUserIdentity

        from scmrepo.git import Stash

        # dulwich will silently generate an empty stash commit if there is
        # nothing to stash, we check status here to get consistent behavior
        # across backends
        if not self.is_dirty(untracked_files=include_untracked):
            return None, False

        if include_untracked or ref == Stash.DEFAULT_STASH:
            # dulwich stash.push does not support include_untracked and does
            # not touch working tree
            raise NotImplementedError

        stash = self._get_stash(ref)
        message_b = message.encode("utf-8") if message else None
        try:
            rev = stash.push(message=message_b)
        except InvalidUserIdentity as exc:
            raise SCMError("Git username and email must be configured") from exc
        return os.fsdecode(rev), True

    def _stash_apply(
        self,
        rev: str,
        reinstate_index: bool = False,
        skip_conflicts: bool = False,
        **kwargs,
    ):
        raise NotImplementedError

    def _stash_drop(self, ref: str, index: int):
        from scmrepo.git import Stash

        if ref == Stash.DEFAULT_STASH:
            raise NotImplementedError

        stash = self._get_stash(ref)
        try:
            stash.drop(index)
        except ValueError as exc:
            raise SCMError("Failed to drop stash entry") from exc

    def _describe(
        self,
        revs: Iterable[str],
        base: Optional[str] = None,
        match: Optional[str] = None,
        exclude: Optional[str] = None,
    ) -> Mapping[str, Optional[str]]:
        if not base:
            base = "refs/tags"
        rev_mapping: dict[str, Optional[str]] = {}
        results: dict[str, Optional[str]] = {}
        for ref in self.iter_refs(base=base):
            if (match and not fnmatch.fnmatch(ref, match)) or (
                exclude and fnmatch.fnmatch(ref, exclude)
            ):
                continue
            revision = self.get_ref(ref, follow=False)
            if revision and revision not in rev_mapping:
                rev_mapping[revision] = ref
        for rev in revs:
            results[rev] = rev_mapping.get(rev)
        return results

    def diff(self, rev_a: str, rev_b: str, binary=False) -> str:
        from dulwich.objects import Commit
        from dulwich.patch import write_tree_diff

        try:
            commit_a = self.repo[os.fsencode(rev_a)]
            commit_b = self.repo[os.fsencode(rev_b)]
        except KeyError as exc:
            raise RevError("Invalid revision") from exc

        assert isinstance(commit_a, Commit)
        assert isinstance(commit_b, Commit)

        buf = BytesIO()
        write_tree_diff(buf, self.repo.object_store, commit_a.tree, commit_b.tree)
        return buf.getvalue().decode("utf-8")

    def reset(self, hard: bool = False, paths: Optional[Iterable[str]] = None):
        raise NotImplementedError

    def checkout_index(
        self,
        paths: Optional[Iterable[str]] = None,
        force: bool = False,
        ours: bool = False,
        theirs: bool = False,
    ):
        raise NotImplementedError

    def status(
        self, ignored: bool = False, untracked_files: str = "all"
    ) -> tuple[Mapping[str, Iterable[str]], Iterable[str], Iterable[str]]:
        from dulwich.porcelain import Error
        from dulwich.porcelain import status as git_status

        with reraise(Error, SCMError("Git status failed")):
            staged, unstaged, untracked = git_status(
                self.repo, ignored=ignored, untracked_files=untracked_files
            )

        return (
            {
                status: [os.fsdecode(name) for name in paths]
                for status, paths in staged.items()
                if paths
            },
            [os.fsdecode(name) for name in unstaged],
            [os.fsdecode(name) for name in untracked],
        )

    def _reset(self) -> None:
        self.__dict__.pop("ignore_manager", None)

    def merge(
        self,
        rev: str,
        commit: bool = True,
        msg: Optional[str] = None,
        squash: bool = False,
    ) -> Optional[str]:
        raise NotImplementedError

    def validate_git_remote(self, url: str, **kwargs):
        from dulwich.client import LocalGitClient, get_transport_and_path
        from dulwich.porcelain import get_remote_repo

        try:
            _, location = get_remote_repo(self.repo, url)
            _config = kwargs.pop("config", StackedConfig.default())
            client, path = get_transport_and_path(location, **kwargs)
        except Exception as exc:
            raise InvalidRemote(url) from exc
        if isinstance(client, LocalGitClient) and not os.path.exists(
            os.path.join("", path)
        ):
            raise InvalidRemote(url)

    def get_remote_url(self, remote: str) -> str:
        from dulwich.porcelain import get_remote_repo

        remote_name, location = get_remote_repo(self.repo, remote)
        if not remote_name:
            raise InvalidRemote(remote)
        return location

    def check_ref_format(self, refname: str) -> bool:
        from dulwich.refs import check_ref_format

        return check_ref_format(refname.encode())

    def get_tag(self, name: str) -> Optional[Union[str, "GitTag"]]:
        from dulwich.objects import Tag

        name_b = os.fsencode(f"refs/tags/{name}")
        try:
            ref = self.repo.refs[name_b]
        except KeyError:
            return None
        if ref in self.repo:
            shafile = self.repo[ref]
            if isinstance(shafile, Tag):
                tag = shafile
                _typ, target_sha = tag.object
                tagger_name, tagger_email = _parse_identity(tag.tagger.decode("utf-8"))
                return GitTag(
                    os.fsdecode(tag.name),
                    tag.id.decode("ascii"),
                    target_sha.decode("ascii"),
                    tagger_name,
                    tagger_email,
                    tag.tag_time,
                    tag.tag_timezone,
                    tag.message.decode("utf-8"),
                )
        return os.fsdecode(ref)

    def get_config(self, path: Optional[str] = None) -> "Config":
        from dulwich.config import ConfigFile

        if path:
            return DulwichConfig(ConfigFile.from_path(path))
        return DulwichConfig(self.repo.get_config_stack())

    def check_attr(
        self,
        path: str,
        attr: str,
        source: Optional[str] = None,
    ) -> Optional[Union[bool, str]]:
        raise NotImplementedError


_IDENTITY_RE = re.compile(r"(?P<name>.+)\s+<(?P<email>.+)>")


def _parse_identity(identity: str) -> tuple[str, str]:
    m = _IDENTITY_RE.match(identity)
    if not m:
        raise SCMError("Could not parse tagger identity '{identity}'")
    return m.group("name"), m.group("email")


def ls_remote(url: str) -> dict[str, Optional[str]]:
    from dulwich import porcelain
    from dulwich.client import HTTPUnauthorized

    try:
        refs = porcelain.ls_remote(url).refs
        return {
            os.fsdecode(ref): sha.decode("ascii") if sha is not None else None
            for ref, sha in refs.items()
        }
    except HTTPUnauthorized as exc:
        raise AuthError(url) from exc
    except Exception as exc:
        raise InvalidRemote(url) from exc



import io
import os
import random

import pytest

from scmrepo.git.credentials import (
    Credential,
    CredentialNotFoundError,
    GitCredentialHelper,
    MemoryCredentialHelper,
)


@pytest.fixture(name="git_helper")
def git_helper_fixture(mocker) -> "GitCredentialHelper":
    mocker.patch("shutil.which", return_value="/usr/bin/git-credential-foo")
    return GitCredentialHelper("foo")


def test_subprocess_get(git_helper, mocker):
    run = mocker.patch(
        "subprocess.run",
        return_value=mocker.Mock(
            stdout="protocol=https\nhost=foo.com\nusername=foo\npassword=bar\n"
        ),
    )
    creds = git_helper.get(Credential(protocol="https", host="foo.com", path="foo.git"))
    assert run.call_args.args[0] == ["git-credential-foo", "get"]
    assert run.call_args.kwargs.get("input") == "protocol=https\nhost=foo.com\n"
    assert creds == Credential(url="https://foo:bar@foo.com")


def test_subprocess_get_use_http_path(git_helper, mocker):
    git_helper.use_http_path = True
    run = mocker.patch(
        "subprocess.run",
        return_value=mocker.Mock(stdout="username=foo\npassword=bar\n"),
    )
    creds = git_helper.get(Credential(protocol="https", host="foo.com", path="foo.git"))
    assert run.call_args.args[0] == ["git-credential-foo", "get"]
    assert (
        run.call_args.kwargs.get("input")
        == "protocol=https\nhost=foo.com\npath=foo.git\n"
    )
    assert creds == Credential(username="foo", password="bar")


def test_subprocess_ignore_unexpected_credential_keys(git_helper, mocker):
    git_helper.use_http_path = True
    run = mocker.patch(
        "subprocess.run",
        # Simulate git-credential-osxkeychain (version >=2.45)
        return_value=mocker.Mock(
            stdout="username=foo\npassword=bar\ncapability[]=state\nstate[]=osxkeychain:seen=1"
        ),
    )
    creds = git_helper.get(Credential(protocol="https", host="foo.com", path="foo.git"))
    assert run.call_args.args[0] == ["git-credential-foo", "get"]
    assert (
        run.call_args.kwargs.get("input")
        == "protocol=https\nhost=foo.com\npath=foo.git\n"
    )
    assert creds == Credential(username="foo", password="bar")


def test_subprocess_strip_trailing_garbage_bytes(git_helper, mocker):
    """Garbage bytes were output from git-credential-osxkeychain from 2.45.0 to 2.47.0
    so must be removed
    https://github.com/git/git/commit/6c3c451fb6e1c3ca83f74e63079d4d0af01b2d69"""
    git_helper.use_http_path = True
    run = mocker.patch(
        "subprocess.run",
        # Simulate git-credential-osxkeychain (version 2.45), assuming initial 0-byte
        return_value=mocker.Mock(
            stdout=f"username=foo\npassword=bar{chr(0)}{random.randbytes(15)}"  # type: ignore[str-bytes-safe]
        ),
    )
    creds = git_helper.get(Credential(protocol="https", host="foo.com", path="foo.git"))
    assert run.call_args.args[0] == ["git-credential-foo", "get"]
    assert (
        run.call_args.kwargs.get("input")
        == "protocol=https\nhost=foo.com\npath=foo.git\n"
    )
    assert creds == Credential(username="foo", password="bar")


def test_subprocess_get_failed(git_helper, mocker):
    from subprocess import CalledProcessError

    mocker.patch("subprocess.run", side_effect=CalledProcessError(1, "/usr/bin/foo"))
    with pytest.raises(CredentialNotFoundError):
        git_helper.get(Credential(protocol="https", host="foo.com"))


def test_subprocess_get_no_output(git_helper, mocker):
    mocker.patch("subprocess.run", return_value=mocker.Mock(stdout="\n"))
    with pytest.raises(CredentialNotFoundError):
        git_helper.get(Credential(protocol="https", host="foo.com"))


def test_subprocess_store(git_helper, mocker):
    run = mocker.patch("subprocess.run")
    git_helper.store(
        Credential(protocol="https", host="foo.com", username="foo", password="bar")
    )
    assert run.call_args.args[0] == ["git-credential-foo", "store"]
    assert (
        run.call_args.kwargs.get("input")
        == "protocol=https\nhost=foo.com\nusername=foo\npassword=bar\n"
    )


def test_subprocess_erase(git_helper, mocker):
    run = mocker.patch("subprocess.run")
    git_helper.erase(Credential(protocol="https", host="foo.com"))
    assert run.call_args.args[0] == ["git-credential-foo", "erase"]
    assert run.call_args.kwargs.get("input") == "protocol=https\nhost=foo.com\n"


def test_memory_helper_get(mocker):
    from getpass import getpass

    from scmrepo.git.credentials import _input_tty

    helper = MemoryCredentialHelper()
    expected = Credential(
        protocol="https", host="foo.com", username="foo", password="bar"
    )
    get_interactive = mocker.patch.object(
        helper,
        "_get_interactive",
        return_value=expected,
    )
    with pytest.raises(CredentialNotFoundError):
        helper.get(Credential(protocol="https", host="foo.com"), interactive=False)
    get_interactive.assert_not_called()
    assert (
        helper.get(Credential(protocol="https", host="foo.com"), interactive=True)
        == expected
    )
    assert get_interactive.call_args.args[0] == Credential(
        protocol="https", host="foo.com"
    )
    assert get_interactive.call_args.args[1:] == (_input_tty, getpass)


def test_memory_helper_get_cached(mocker):
    helper = MemoryCredentialHelper()
    expected = Credential(
        protocol="https", host="foo.com", username="foo", password="bar"
    )
    helper[expected] = expected

    get_interactive = mocker.patch.object(
        helper,
        "_get_interactive",
        return_value=expected,
    )
    assert (
        helper.get(Credential(protocol="https", host="foo.com"), interactive=False)
        == expected
    )
    get_interactive.assert_not_called()


def test_memory_helper_prompt_disabled(mocker):
    helper = MemoryCredentialHelper()
    get_interactive = mocker.patch.object(
        helper,
        "_get_interactive",
    )
    mocker.patch.dict(os.environ, {"GIT_TERMINAL_PROMPT": "0"})
    with pytest.raises(CredentialNotFoundError):
        helper.get(Credential(protocol="https", host="foo.com"), interactive=True)
    get_interactive.assert_not_called()


def test_memory_helper_prompt_askpass(mocker):
    helper = MemoryCredentialHelper()
    mocker.patch.dict(os.environ, {"GIT_ASKPASS": "/usr/local/bin/my-askpass"})
    run = mocker.patch(
        "subprocess.run",
        side_effect=[
            mocker.Mock(stdout="foo"),
            mocker.Mock(stdout="bar\n"),
        ],
    )
    expected = Credential(
        protocol="https", host="foo.com", username="foo", password="bar"
    )
    assert (
        helper.get(Credential(protocol="https", host="foo.com"), interactive=True)
        == expected
    )
    assert len(run.call_args_list) == 2
    assert run.call_args_list[0].args[0] == [
        "/usr/local/bin/my-askpass",
        "Username for 'https://foo.com': ",
    ]
    assert run.call_args_list[1].args[0] == [
        "/usr/local/bin/my-askpass",
        "Password for 'https://foo@foo.com': ",
    ]


def test_get_matching_commands():
    from dulwich.config import ConfigFile

    config_file = io.BytesIO(
        b"""
[credential]
    helper = /usr/local/bin/my-helper
    UseHttpPath = true
"""
    )
    config_file.seek(0)
    config = ConfigFile.from_file(config_file)
    assert list(
        GitCredentialHelper.get_matching_commands("https://foo.com/foo.git", config)
    ) == [("/usr/local/bin/my-helper", True)]

    config_file = io.BytesIO(
        b"""
[credential]
    helper = /usr/local/bin/my-helper
"""
    )
    config_file.seek(0)
    config = ConfigFile.from_file(config_file)
    assert list(
        GitCredentialHelper.get_matching_commands("https://foo.com/foo.git", config)
    ) == [("/usr/local/bin/my-helper", False)]

    config_file = io.BytesIO(
        b"""
[credential]
    helper =
"""
    )
    config_file.seek(0)
    config = ConfigFile.from_file(config_file)
    assert (
        list(
            GitCredentialHelper.get_matching_commands("https://foo.com/foo.git", config)
        )
        == []
    )



"""git config convenience wrapper."""

import logging
from abc import ABC, abstractmethod
from collections.abc import Iterator

logger = logging.getLogger(__name__)


class Config(ABC):
    """Read-only Git config."""

    @abstractmethod
    def get(self, section: tuple[str, ...], name: str) -> str:
        """Return the specified setting as a string.

        Raises:
            KeyError: Option was not set.
        """

    @abstractmethod
    def get_bool(self, section: tuple[str, ...], name: str) -> bool:
        """Return the specified setting as a boolean.

        Raises:
            KeyError: Option was not set.
            ValueError: Option is not a valid boolean.
        """

    @abstractmethod
    def get_multivar(self, section: tuple[str, ...], name: str) -> Iterator[str]:
        """Iterate over string values in the specified multivar setting.

        Raises:
            KeyError: Option was not set.
        """



import logging
import sys
from typing import Any, BinaryIO, Callable, ClassVar, Optional, Union

from fsspec.callbacks import DEFAULT_CALLBACK, Callback, TqdmCallback
from tqdm import tqdm

from scmrepo.progress import GitProgressEvent


class _Tqdm(tqdm):
    """
    maximum-compatibility tqdm-based progressbars
    """

    BAR_FMT_DEFAULT = (
        "{percentage:3.0f}% {desc}|{bar}|"
        "{postfix[info]}{n_fmt}/{total_fmt}"
        " [{elapsed}<{remaining}, {rate_fmt:>11}]"
    )
    # nested bars should have fixed bar widths to align nicely
    BAR_FMT_DEFAULT_NESTED = (
        "{percentage:3.0f}%|{bar:10}|{desc:{ncols_desc}.{ncols_desc}}"
        "{postfix[info]}{n_fmt}/{total_fmt}"
        " [{elapsed}<{remaining}, {rate_fmt:>11}]"
    )
    BAR_FMT_NOTOTAL = "{desc}{bar:b}|{postfix[info]}{n_fmt} [{elapsed}, {rate_fmt:>11}]"
    BYTES_DEFAULTS: ClassVar[dict[str, Any]] = {
        "unit": "B",
        "unit_scale": True,
        "unit_divisor": 1024,
        "miniters": 1,
    }

    def __init__(
        self,
        iterable=None,
        disable=None,
        level=logging.ERROR,
        desc=None,
        leave=False,
        bar_format=None,
        bytes=False,  # noqa: A002
        file=None,
        total=None,
        postfix=None,
        **kwargs,
    ):
        kwargs = kwargs.copy()
        if bytes:
            kwargs = {**self.BYTES_DEFAULTS, **kwargs}
        else:
            kwargs.setdefault("unit_scale", total > 999 if total else True)
        if file is None:
            file = sys.stderr
        super().__init__(
            iterable=iterable,
            disable=disable,
            leave=leave,
            desc=desc,
            bar_format="!",
            lock_args=(False,),
            total=total,
            **kwargs,
        )
        self.postfix = postfix or {"info": ""}
        if bar_format is None:
            if self.__len__():
                self.bar_format = (
                    self.BAR_FMT_DEFAULT_NESTED if self.pos else self.BAR_FMT_DEFAULT
                )
            else:
                self.bar_format = self.BAR_FMT_NOTOTAL
        else:
            self.bar_format = bar_format
        self.refresh()

    def update_to(self, current, total=None):
        if total:
            self.total = total
        self.update(current - self.n)

    def close(self):
        self.postfix["info"] = ""
        # remove ETA (either unknown or zero); remove completed bar
        self.bar_format = self.bar_format.replace("<{remaining}", "").replace(
            "|{bar:10}|", " "
        )
        super().close()

    @property
    def format_dict(self):
        """inject `ncols_desc` to fill the display width (`ncols`)"""
        d = super().format_dict
        ncols = d["ncols"] or 80
        # assumes `bar_format` has max one of ("ncols_desc" & "ncols_info")

        meter = self.format_meter(  # type: ignore[call-arg]
            ncols_desc=1, ncols_info=1, **d
        )
        ncols_left = ncols - len(meter) + 1
        ncols_left = max(ncols_left, 0)
        if ncols_left:
            d["ncols_desc"] = d["ncols_info"] = ncols_left
        else:
            # work-around for zero-width description
            d["ncols_desc"] = d["ncols_info"] = 1
            d["prefix"] = ""
        return d


class LFSCallback(Callback):
    """Callback subclass to generate Git/LFS style progress."""

    def __init__(
        self,
        *args,
        git_progress: Optional[Callable[[GitProgressEvent], None]] = None,
        direction: str = "Downloading",
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.direction = direction
        self.git_progress = git_progress

    def call(self, *args, **kwargs):
        super().call(*args, **kwargs)
        self._update_git()

    def _update_git(self):
        if not self.git_progress:
            return
        event = GitProgressEvent(
            phase=f"{self.direction} LFS objects",
            completed=self.value,
            total=self.size,
        )
        self.git_progress(event)

    def branched(self, path_1: Union[str, BinaryIO], path_2: str, **kwargs):
        if self.git_progress:
            return TqdmCallback(
                tqdm_kwargs={
                    "desc": path_1 if isinstance(path_1, str) else path_2,
                    "bytes": True,
                },
                tqdm_cls=_Tqdm,
            )
        return DEFAULT_CALLBACK

    @classmethod
    def as_lfs_callback(
        cls,
        git_progress: Optional[Callable[[GitProgressEvent], None]] = None,
        **kwargs,
    ):
        if git_progress is None:
            return DEFAULT_CALLBACK
        return cls(git_progress=git_progress, **kwargs)



import io
import locale
import logging
import os
import re
import sys
from collections.abc import Iterable, Mapping
from functools import partial, wraps
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Optional,
    Union,
)

from funcy import ignore

from scmrepo.exceptions import (
    CloneError,
    InvalidRemote,
    MergeConflictError,
    RevError,
    SCMError,
    UnsupportedIndexFormat,
)
from scmrepo.git.objects import GitCommit, GitObject, GitTag
from scmrepo.utils import relpath

from .base import BaseGitBackend, SyncStatus

if TYPE_CHECKING:
    from scmrepo.git.config import Config
    from scmrepo.progress import GitProgressEvent


logger = logging.getLogger(__name__)


def requires_git(f: Callable):
    @wraps(f)
    def wrapper(*args, **kwargs):
        try:
            from git.exc import GitCommandNotFound
        except ImportError as exc:
            raise NotImplementedError from exc

        try:
            return f(*args, **kwargs)
        except GitCommandNotFound as exc:
            raise NotImplementedError from exc

    return wrapper


# NOTE: Check if we are in a bundle
# https://pythonhosted.org/PyInstaller/runtime-information.html
def is_binary() -> bool:
    return getattr(sys, "frozen", False)


def fix_env(env: Optional[dict[str, str]] = None) -> dict[str, str]:
    if env is None:
        environ = os.environ.copy()
    else:
        environ = env.copy()

    if is_binary():
        lp_key = "LD_LIBRARY_PATH"
        lp_orig = environ.get(lp_key + "_ORIG", None)
        if lp_orig is not None:
            environ[lp_key] = lp_orig
        else:
            environ.pop(lp_key, None)
    return environ


class GitPythonObject(GitObject):
    def __init__(self, obj):
        self.obj = obj

    def open(
        self,
        mode: str = "r",
        encoding: Optional[str] = None,
        raw: bool = True,
        **kwargs,
    ):
        if not raw:
            raise NotImplementedError
        if not encoding:
            encoding = locale.getpreferredencoding(False)
        # GitPython's obj.data_stream is a fragile thing, it is better to
        # read it immediately, also it needs to be to decoded if we follow
        # the `open()` behavior (since data_stream.read() returns bytes,
        # and `open` with default "r" mode returns str)
        data = self.obj.data_stream.read()
        if mode == "rb":
            return io.BytesIO(data)
        return io.StringIO(data.decode(encoding))

    @property
    def name(self) -> str:
        # NOTE: `obj.name` is not always a basename. See [1] for more details.
        #
        # [1] https://github.com/treeverse/dvc/issues/3481
        return os.path.basename(self.obj.path)

    @property
    def mode(self) -> int:
        return self.obj.mode

    def scandir(self) -> Iterable["GitPythonObject"]:
        for obj in self.obj:
            yield GitPythonObject(obj)

    @property
    def size(self) -> int:
        return self.obj.size

    @property
    def sha(self) -> str:
        return self.obj.hexsha


class GitPythonBackend(BaseGitBackend):  # pylint:disable=abstract-method
    """git-python Git backend."""

    @requires_git
    def __init__(  # pylint:disable=W0231
        self, root_dir=os.curdir, search_parent_directories=True
    ):
        import git
        from git.exc import InvalidGitRepositoryError

        try:
            self.repo = git.Repo(
                root_dir, search_parent_directories=search_parent_directories
            )
        except InvalidGitRepositoryError as exc:
            msg = "{} is not a git repository"
            raise SCMError(msg.format(root_dir)) from exc

        # NOTE: fixing LD_LIBRARY_PATH for binary built by PyInstaller.
        # http://pyinstaller.readthedocs.io/en/stable/runtime-information.html
        env = fix_env()
        libpath = env.get("LD_LIBRARY_PATH", None)
        self.repo.git.update_environment(LD_LIBRARY_PATH=libpath)

    def close(self):
        self.repo.close()

    @property
    def git(self):
        return self.repo.git

    def is_ignored(self, path: "Union[str, os.PathLike[str]]") -> bool:
        from git.exc import GitCommandError

        func = ignore(GitCommandError)(self.repo.git.check_ignore)
        return bool(func(str(path)))

    @property
    def root_dir(self) -> Optional[str]:
        d = self.repo.working_tree_dir
        return os.fspath(d) if d is not None else d

    @staticmethod
    @requires_git
    def clone(
        url: str,
        to_path: Union[str, os.PathLike[str]],
        shallow_branch: Optional[str] = None,
        progress: Optional[Callable[["GitProgressEvent"], None]] = None,
        bare: bool = False,
        mirror: bool = False,
    ):
        from git import Repo
        from git.exc import GitCommandError

        ld_key = "LD_LIBRARY_PATH"

        env = fix_env()
        if is_binary() and ld_key not in env:
            # In fix_env, we delete LD_LIBRARY_PATH key if it was empty before
            # PyInstaller modified it. GitPython, in git.Repo.clone_from, uses
            # env to update its own internal state. When there is no key in
            # env, this value is not updated and GitPython reuses
            # LD_LIBRARY_PATH that has been set by PyInstaller.
            # See [1] for more info.
            # [1] https://github.com/gitpython-developers/GitPython/issues/924
            env[ld_key] = ""

        try:
            if shallow_branch is not None and os.path.exists(url):
                # git disables --depth for local clones unless file:// url
                # scheme is used
                url = f"file://{url}"

            from scmrepo.progress import GitProgressReporter

            clone_from = partial(
                Repo.clone_from,
                url,
                to_path,
                env=env,  # needed before we can fix it in __init__
                no_single_branch=True,
                progress=GitProgressReporter.wrap_fn(progress) if progress else None,
                bare=bare,
                mirror=mirror,
            )
            if shallow_branch is None:
                tmp_repo = clone_from()
            else:
                tmp_repo = clone_from(branch=shallow_branch, depth=1)
            tmp_repo.close()
        except GitCommandError as exc:  # pylint: disable=no-member
            raise CloneError(url, os.fsdecode(to_path)) from exc

    @staticmethod
    @requires_git
    def init(path: str, bare: bool = False) -> None:
        from funcy import retry
        from git import Repo
        from git.exc import GitCommandNotFound

        # NOTE: handles EAGAIN error on BSD systems (osx in our case).
        # Otherwise when running tests you might get this exception:
        #
        #    GitCommandNotFound: Cmd('git') not found due to:
        #        OSError('[Errno 35] Resource temporarily unavailable')
        method = retry(5, GitCommandNotFound)(Repo.init)
        git = method(path, bare=bare)
        git.close()

    @staticmethod
    @requires_git
    def is_sha(rev):
        import git

        return rev and git.Repo.re_hexsha_shortened.search(rev)

    @property
    def dir(self) -> str:
        return os.fspath(self.repo.git_dir)

    def add(
        self,
        paths: Union[str, Iterable[str]],
        update: bool = False,
        force: bool = False,
    ):
        try:
            if update or not force:
                # NOTE: git-python index.add() defines force parameter but
                # ignores it (index.add() behavior is always force=True)
                kwargs: dict[str, Any] = {}
                if update:
                    kwargs["update"] = True
                if isinstance(paths, str):
                    paths = [paths]
                if not force:
                    paths = [path for path in paths if not self.is_ignored(path)]
                self.git.add(*paths, **kwargs)
            else:
                self.repo.index.add(paths if isinstance(paths, str) else list(paths))
        except AssertionError as exc:
            # NOTE: GitPython is not currently able to handle index version >= 3.
            # See https://github.com/treeverse/dvc/issues/610 for more details.
            raise UnsupportedIndexFormat from exc

    def commit(self, msg: str, no_verify: bool = False):
        from git.exc import HookExecutionError

        try:
            self.repo.index.commit(msg, skip_hooks=no_verify)
        except HookExecutionError as exc:
            raise SCMError("Git pre-commit hook failed") from exc

    def checkout(
        self,
        branch: str,
        create_new: Optional[bool] = False,
        force: bool = False,
        **kwargs,
    ):
        if create_new:
            self.repo.git.checkout("HEAD", b=branch, force=force, **kwargs)
        else:
            self.repo.git.checkout(branch, force=force, **kwargs)

    def fetch(
        self,
        remote: Optional[str] = None,
        force: bool = False,
        unshallow: bool = False,
    ):
        if not remote:
            remote = "origin"
        kwargs = {}
        if force:
            kwargs["force"] = True
        if unshallow:
            kwargs["unshallow"] = True
        infos = self.repo.remote(name=remote).fetch(**kwargs)  # type: ignore[arg-type]
        for info in infos:
            if info.flags & info.ERROR:
                raise SCMError(f"fetch failed: {info.note}")

    def pull(self, **kwargs):
        infos = self.repo.remote().pull(**kwargs)
        for info in infos:
            if info.flags & info.ERROR:
                raise SCMError(f"pull failed: {info.note}")

    def push(self):
        infos = self.repo.remote().push()
        for info in infos:
            if info.flags & info.ERROR:
                raise SCMError(f"push failed: {info.summary}")

    def branch(self, branch):
        self.repo.git.branch(branch)

    def tag(
        self,
        tag: str,
        target: Optional[str] = None,
        annotated: bool = False,
        message: Optional[str] = None,
    ):
        if annotated and not message:
            raise SCMError("message is required for annotated tag")
        self.repo.git.tag(tag, target or "HEAD", a=annotated, m=message)

    def untracked_files(self):
        files = self.repo.untracked_files
        return [os.path.join(self.repo.working_dir, fname) for fname in files]

    def is_tracked(self, path):
        return bool(self.repo.git.ls_files(path))

    def is_dirty(self, untracked_files: bool = False) -> bool:
        return self.repo.is_dirty(untracked_files=untracked_files)

    def active_branch(self):
        try:
            return self.repo.active_branch.name
        except TypeError as exc:
            raise SCMError("No active branch") from exc

    def active_branch_remote(self) -> str:
        try:
            return self.repo.active_branch.tracking_branch()  # type: ignore[return-value]
        except (TypeError, ValueError) as exc:
            raise SCMError("No active branch tracking remote") from exc

    def list_branches(self):
        return [h.name for h in self.repo.heads]

    def list_tags(self):
        return [t.name for t in self.repo.tags]

    def list_all_commits(self):
        head = self.get_ref("HEAD")
        if not head:
            # Empty repo
            return []

        return [
            c.hexsha
            for c in self.repo.iter_commits(
                rev=head, branches=True, tags=True, remotes=True
            )
        ]

    def get_tree_obj(self, rev: str, **kwargs) -> GitPythonObject:
        tree = self.repo.tree(rev)
        return GitPythonObject(tree)

    def get_rev(self):
        return self.repo.rev_parse("HEAD").hexsha

    def resolve_rev(self, rev):
        from contextlib import suppress

        from git.exc import BadName, GitCommandError

        def _resolve_rev(name):
            with suppress(BadName, GitCommandError):
                try:
                    # Try python implementation of rev-parse first, it's faster
                    return self.repo.rev_parse(name).hexsha
                except NotImplementedError:
                    # Fall back to `git rev-parse` for advanced features
                    return self.repo.git.rev_parse(name)
                except ValueError as exc:
                    raise RevError(f"unknown Git revision '{name}'") from exc

        # Resolve across local names
        sha = _resolve_rev(rev)
        if sha:
            return sha

        # Try all the remotes and if it resolves unambiguously then take it
        if not self.is_sha(rev):
            shas = {
                _resolve_rev(f"{remote.name}/{rev}") for remote in self.repo.remotes
            } - {None}
            if len(shas) > 1:
                raise RevError(f"ambiguous Git revision '{rev}'")
            if len(shas) == 1:
                return shas.pop()

        raise RevError(f"unknown Git revision '{rev}'")

    def resolve_commit(self, rev: str) -> "GitCommit":
        """Return Commit object for the specified revision."""
        from git.exc import BadName, GitCommandError
        from git.objects.tag import TagObject
        from gitdb.exc import BadObject

        try:
            commit = self.repo.rev_parse(rev)
        except (BadName, BadObject, GitCommandError) as exc:
            raise SCMError(f"Invalid commit '{rev}'") from exc
        if isinstance(commit, TagObject):
            commit = commit.object
        return GitCommit(
            commit.hexsha,
            commit.committed_date,
            commit.committer_tz_offset,  # type: ignore[arg-type]
            commit.message,  # type: ignore[arg-type]
            [str(parent) for parent in commit.parents],
            commit.committer.name,  # type: ignore[arg-type]
            commit.committer.email,  # type: ignore[arg-type]
            commit.author.name,  # type: ignore[arg-type]
            commit.author.email,  # type: ignore[arg-type]
            commit.authored_date,
            commit.author_tz_offset,  # type: ignore[arg-type]
        )

    def set_ref(
        self,
        name: str,
        new_ref: str,
        old_ref: Optional[str] = None,
        message: Optional[str] = None,
        symbolic: Optional[bool] = False,
    ):
        from git.exc import GitCommandError

        if old_ref and self.get_ref(name) != old_ref:
            raise SCMError(f"Failed to set ref '{name}'")
        try:
            if symbolic:
                if message:
                    self.git.symbolic_ref(name, new_ref, m=message)
                else:
                    self.git.symbolic_ref(name, new_ref)
            else:
                args = [name, new_ref]
                if old_ref:
                    args.append(old_ref)
                if message:
                    self.git.update_ref(*args, m=message, create_reflog=True)
                else:
                    self.git.update_ref(*args)
        except GitCommandError as exc:
            raise SCMError(f"Failed to set ref '{name}'") from exc

    def get_ref(self, name: str, follow: bool = True) -> Optional[str]:  # noqa: C901, PLR0911
        from git.exc import GitCommandError

        if name == "HEAD":
            try:
                if follow or self.repo.head.is_detached:
                    return self.repo.head.commit.hexsha
                return f"refs/heads/{self.repo.active_branch}"
            except (GitCommandError, ValueError):
                return None
        elif name.startswith("refs/heads/"):
            name = name[11:]
            if name in self.repo.heads:
                return self.repo.heads[name].commit.hexsha
        elif name.startswith("refs/tags/"):
            name = name[10:]
            if name in self.repo.tags:
                return self.repo.tags[name].commit.hexsha
        else:
            if not follow:
                try:
                    rev = self.git.symbolic_ref(name).strip()
                    return rev if rev else None
                except GitCommandError:
                    pass
            try:
                rev = self.git.show_ref(name, hash=True).strip()
                return rev if rev else None
            except GitCommandError:
                pass
        return None

    def remove_ref(self, name: str, old_ref: Optional[str] = None):
        from git.exc import GitCommandError

        if old_ref and self.get_ref(name) != old_ref:
            raise SCMError(f"Failed to set ref '{name}'")
        try:
            args = [name]
            if old_ref:
                args.append(old_ref)
            self.git.update_ref(*args, d=True)
        except GitCommandError as exc:
            raise SCMError(f"Failed to set ref '{name}'") from exc

    def iter_refs(self, base: Optional[str] = None):
        from git import Reference

        for ref in Reference.iter_items(self.repo, common_path=base):
            yield ref.path

    def iter_remote_refs(self, url: str, base: Optional[str] = None, **kwargs):
        raise NotImplementedError

    def get_refs_containing(self, rev: str, pattern: Optional[str] = None):
        from git.exc import GitCommandError

        try:
            if pattern:
                args = [pattern]
            else:
                args = []
            for line in self.git.for_each_ref(
                *args, contains=rev, format=r"%(refname)"
            ).splitlines():
                line = line.strip()
                if line:
                    yield line
        except GitCommandError:
            pass

    def push_refspecs(
        self,
        url: str,
        refspecs: Union[str, Iterable[str]],
        force: bool = False,
        on_diverged: Optional[Callable[[str, str], bool]] = None,
        progress: Optional[Callable[["GitProgressEvent"], None]] = None,
        **kwargs,
    ) -> Mapping[str, SyncStatus]:
        raise NotImplementedError

    def fetch_refspecs(
        self,
        url: str,
        refspecs: Union[str, Iterable[str]],
        force: bool = False,
        on_diverged: Optional[Callable[[str, str], bool]] = None,
        progress: Optional[Callable[["GitProgressEvent"], None]] = None,
        **kwargs,
    ) -> Mapping[str, SyncStatus]:
        raise NotImplementedError

    def _stash_iter(self, ref: str):
        raise NotImplementedError

    def _stash_push(
        self,
        ref: str,
        message: Optional[str] = None,
        include_untracked: bool = False,
    ) -> tuple[Optional[str], bool]:
        from scmrepo.git import Stash

        if not self.is_dirty(untracked_files=include_untracked):
            return None, False

        args = ["push"]
        if message:
            args.extend(["-m", message])
        if include_untracked:
            args.append("--include-untracked")
        self.git.stash(*args)
        commit = self.resolve_commit("stash@{0}")
        if ref != Stash.DEFAULT_STASH:
            # `git stash` CLI doesn't support using custom refspecs,
            # so we push a commit onto refs/stash, make our refspec
            # point to the new commit, then pop it from refs/stash
            # `git stash create` is intended to be used for this kind of
            # behavior but it doesn't support --include-untracked so we need to
            # use push
            self.set_ref(ref, commit.hexsha, message=commit.message)
            self.git.stash("drop")
        return commit.hexsha, False

    def _stash_apply(
        self,
        rev: str,
        reinstate_index: bool = False,
        skip_conflicts: bool = False,
        **kwargs,
    ):
        from git.exc import GitCommandError

        if skip_conflicts:
            raise NotImplementedError
        try:
            args = ["apply"]
            if reinstate_index:
                args.append("--index")
            args.append(rev)
            self.git.stash(args)
        except GitCommandError as exc:
            out = str(exc)
            if "CONFLICT" in out or "already exists" in out:
                raise MergeConflictError(
                    "Stash apply resulted in merge conflicts"
                ) from exc
            raise SCMError("Could not apply stash") from exc

    def _stash_drop(self, ref: str, index: int):
        from git.exc import GitCommandError

        from scmrepo.git import Stash

        if ref == Stash.DEFAULT_STASH:
            self.git.stash("drop", index)
            return

        self.git.reflog("delete", "--updateref", "--rewrite", f"{ref}@{{{index}}}")
        try:
            self.git.reflog("exists", ref)
        except GitCommandError:
            self.remove_ref(ref)

    def _describe(
        self,
        revs: Iterable[str],
        base: Optional[str] = None,
        match: Optional[str] = None,
        exclude: Optional[str] = None,
    ) -> Mapping[str, Optional[str]]:
        raise NotImplementedError

    def diff(self, rev_a: str, rev_b: str, binary=False) -> str:
        raise NotImplementedError

    def reset(self, hard: bool = False, paths: Optional[Iterable[str]] = None):
        if paths:
            paths_list: Optional[list[str]] = [
                relpath(path, self.root_dir) for path in paths
            ]
            if os.name == "nt":
                paths_list = [
                    path.replace("\\", "/")
                    for path in paths_list  # type: ignore[union-attr]
                ]
        else:
            paths_list = None
        self.repo.head.reset(index=True, working_tree=hard, paths=paths_list)

    def checkout_index(
        self,
        paths: Optional[Iterable[str]] = None,
        force: bool = False,
        ours: bool = False,
        theirs: bool = False,
    ):
        """Checkout the specified paths from HEAD index."""
        assert not (ours and theirs)
        if ours or theirs:
            args = ["--ours"] if ours else ["--theirs"]
            if force:
                args.append("--force")
            args.append("--")
            if paths:
                args.extend(list(paths))
            else:
                args.append(".")
            self.repo.git.checkout(*args)
        else:
            if paths:
                paths_list: Optional[list[str]] = [
                    relpath(path, self.root_dir) for path in paths
                ]
                if os.name == "nt":
                    paths_list = [
                        path.replace("\\", "/")
                        for path in paths_list  # type: ignore[union-attr]
                    ]
            else:
                paths_list = None
            self.repo.index.checkout(paths=paths_list, force=force)

    def status(
        self, ignored: bool = False, untracked_files: str = "all"
    ) -> tuple[Mapping[str, Iterable[str]], Iterable[str], Iterable[str]]:
        raise NotImplementedError

    def merge(
        self,
        rev: str,
        commit: bool = True,
        msg: Optional[str] = None,
        squash: bool = False,
    ) -> Optional[str]:
        from git.exc import GitCommandError

        if commit and squash:
            raise SCMError("Cannot merge with 'squash' and 'commit'")

        if commit and not msg:
            raise SCMError("Merge commit message is required")

        merge = partial(self.git.merge, rev)
        try:
            if commit:
                merge(m=msg)
                return self.get_rev()
            merge(no_commit=True, squash=True)
        except GitCommandError as exc:
            if "CONFLICT" in str(exc):
                raise MergeConflictError("Merge contained conflicts") from exc
            raise SCMError("Merge failed") from exc
        return None

    def validate_git_remote(self, url: str, **kwargs):
        raise NotImplementedError

    def get_remote_url(self, remote: str) -> str:
        from git.exc import GitCommandError

        try:
            return self.repo.remotes[remote].url
        except (KeyError, GitCommandError) as exc:
            raise InvalidRemote(remote) from exc

    def check_ref_format(self, refname: str):
        raise NotImplementedError

    def get_tag(self, name: str) -> Optional[Union[str, "GitTag"]]:
        try:
            ref = self.repo.tags[name]
            if not ref.tag:
                return ref.commit.hexsha
            tag = ref.tag
            assert tag.tagger.email
            assert tag.tagger.name
            return GitTag(
                tag.tag,
                tag.hexsha,
                tag.object.hexsha,
                tag.tagger.name,
                tag.tagger.email,
                tag.tagged_date,
                tag.tagger_tz_offset,
                tag.message,
            )
        except IndexError:
            pass
        return None

    def get_config(self, path: Optional[str] = None) -> "Config":
        raise NotImplementedError

    def check_attr(
        self,
        path: str,
        attr: str,
        source: Optional[str] = None,
    ) -> Optional[Union[bool, str]]:
        from git.exc import GitCommandError

        try:
            result = self.git.check_attr(attr, "--", path, source=source)
        except GitCommandError as exc:
            raise SCMError("Failed to check attribute") from exc
        escaped_path = re.escape(path)
        escaped_attr = re.escape(attr)
        m = re.match(f"{escaped_path}: {escaped_attr}: (?P<info>.*)", result)
        if not m or not m.group("info"):
            raise SCMError("Failed to check attribute")
        info = m.group("info")
        if info == "unspecified":
            return None
        if info == "set":
            return True
        if info == "unset":
            return False
        return info

    def list_submodules(self) -> list[str]:
        raise NotImplementedError



from dataclasses import dataclass, fields
from typing import Any


@dataclass(frozen=True)
class LFSObject:
    oid: str
    size: int

    def __str__(self) -> str:
        return self.oid

    @classmethod
    def from_dict(cls, d: dict[str, Any]) -> "LFSObject":
        return cls(**{k: v for k, v in d.items() if k in fields(cls)})



import errno
import os
from collections.abc import Collection
from typing import TYPE_CHECKING, BinaryIO, Callable, Optional, Union

from .pointer import Pointer
from .progress import LFSCallback

if TYPE_CHECKING:
    from scmrepo.git import Git
    from scmrepo.progress import GitProgressEvent


class LFSStorage:
    def __init__(self, path: Union[str, os.PathLike[str]]):
        self.path = path

    def fetch(
        self,
        url: str,
        objects: Collection[Pointer],
        progress: Optional[Callable[["GitProgressEvent"], None]] = None,
        batch_size: Optional[int] = None,
    ):
        from .client import LFSClient

        with LFSCallback.as_lfs_callback(progress) as cb:
            cb.set_size(len(objects))
            with LFSClient.from_git_url(url) as client:
                client.download(self, objects, callback=cb, batch_size=batch_size)

    def oid_to_path(self, oid: str):
        return os.path.join(self.path, "objects", oid[0:2], oid[2:4], oid)

    def exists(self, obj: Union[Pointer, str]):
        oid = obj if isinstance(obj, str) else obj.oid
        path = self.oid_to_path(oid)
        return os.path.exists(path)

    def open(
        self,
        obj: Union[Pointer, str],
        fetch_url: Optional[str] = None,
        batch_size: Optional[int] = None,
        **kwargs,
    ) -> BinaryIO:
        oid = obj if isinstance(obj, str) else obj.oid
        path = self.oid_to_path(oid)
        try:
            return open(path, **kwargs)
        except FileNotFoundError:
            if not fetch_url or not isinstance(obj, Pointer):
                raise
        try:
            self.fetch(fetch_url, [obj], batch_size=batch_size)
        except BaseException as exc:
            raise FileNotFoundError(
                errno.ENOENT, os.strerror(errno.ENOENT), path
            ) from exc
        return open(path, **kwargs)

    def close(self):
        pass


def get_storage_path(scm: "Git") -> str:
    """Return the LFS storage directory for the specified repository."""

    config = scm.get_config()
    git_dir = scm._get_git_dir(scm.root_dir)  # pylint: disable=protected-access
    try:
        path = config.get(("lfs",), "storage")
        if os.path.isabs(path):
            return path
    except KeyError:
        path = "lfs"
    return os.path.join(git_dir, path)



import os

import pytest
from pytest_test_utils import TmpDir

from scmrepo.git import Git


@pytest.fixture(name="make_fs")
def fixture_make_fs(scm: Git, git: Git):
    def _make_fs(rev=None):
        from scmrepo.fs import GitFileSystem
        from scmrepo.git.objects import GitTrie

        # NOTE: not all git backends have `resolve_rev` implemented,
        # so we are using whichever works.
        resolved = scm.resolve_rev(rev or "HEAD")
        tree = git.get_tree_obj(rev=resolved)
        trie = GitTrie(tree, resolved)
        return GitFileSystem(trie=trie)

    return _make_fs


@pytest.mark.parametrize("raw", [True, False])
def test_open(tmp_dir: TmpDir, scm: Git, make_fs, raw: bool, git_backend: str):
    if not raw and git_backend != "pygit2":
        pytest.skip()

    files = tmp_dir.gen({"foo": "foo", "тест": "проверка", "data": {"lorem": "ipsum"}})
    scm.add_commit(files, message="add")

    fs = make_fs()
    with fs.open("foo", mode="r", encoding="utf-8", raw=raw) as fobj:
        assert fobj.read() == "foo"
    with fs.open("тест", mode="r", encoding="utf-8", raw=raw) as fobj:
        assert fobj.read() == "проверка"
    with pytest.raises(IOError):  # noqa: PT011
        fs.open("not-existing-file", raw=raw)
    with pytest.raises(IOError):  # noqa: PT011
        fs.open("data", raw=raw)


def test_exists(tmp_dir: TmpDir, scm: Git, make_fs):
    scm.commit("init")
    files = tmp_dir.gen({"foo": "foo", "тест": "проверка", "data": {"lorem": "ipsum"}})

    fs = make_fs()

    assert fs.exists("/")
    assert fs.exists(".")
    assert not fs.exists("foo")
    assert not fs.exists("тест")
    assert not fs.exists("data")
    assert not fs.exists("data/lorem")

    scm.add_commit(files, message="add")

    fs = make_fs()
    assert fs.exists("/")
    assert fs.exists(".")
    assert fs.exists("foo")
    assert fs.exists("тест")
    assert fs.exists("data")
    assert fs.exists("data/lorem")
    assert not fs.exists("non-existing-file")


def test_isdir(tmp_dir: TmpDir, scm: Git, make_fs):
    tmp_dir.gen({"foo": "foo", "тест": "проверка", "data": {"lorem": "ipsum"}})
    scm.add_commit(["foo", "data"], message="add")

    fs = make_fs()

    assert fs.isdir("/")
    assert fs.isdir(".")
    assert fs.isdir("data")
    assert not fs.isdir("foo")
    assert not fs.isdir("non-existing-file")


def test_isfile(tmp_dir: TmpDir, scm: Git, make_fs):
    tmp_dir.gen({"foo": "foo", "тест": "проверка", "data": {"lorem": "ipsum"}})
    scm.add_commit(["foo", "data"], message="add")

    fs = make_fs()
    assert not fs.isfile("/")
    assert not fs.isfile(".")
    assert fs.isfile("foo")
    assert not fs.isfile("data")
    assert not fs.isfile("not-existing-file")


def test_walk(tmp_dir: TmpDir, scm: Git, make_fs):
    tmp_dir.gen(
        {
            "foo": "foo",
            "тест": "проверка",
            "data": {"lorem": "ipsum", "subdir": {"sub": "sub"}},
        }
    )
    scm.add_commit("data/subdir", message="add")
    fs = make_fs()

    def convert_to_sets(walk_results):
        return [(root, set(dirs), set(nondirs)) for root, dirs, nondirs in walk_results]

    assert convert_to_sets(fs.walk("/")) == convert_to_sets(
        [
            ("/", ["data"], []),
            ("/data", ["subdir"], []),
            (
                "/data/subdir",
                [],
                ["sub"],
            ),
        ]
    )

    assert convert_to_sets(fs.walk("data/subdir")) == convert_to_sets(
        [
            (
                "data/subdir",
                [],
                ["sub"],
            )
        ]
    )


def test_walk_with_submodules(
    scm: Git,
    remote_git_dir: TmpDir,
    make_fs,
):
    remote_git = Git(remote_git_dir)
    remote_git_dir.gen({"foo": "foo", "bar": "bar", "dir": {"data": "data"}})
    remote_git.add_commit(["foo", "bar", "dir"], message="add dir and files")
    scm.gitpython.repo.create_submodule(
        "submodule", "submodule", url=os.fspath(remote_git_dir)
    )
    scm.commit("added submodule")

    files = []
    dirs = []
    fs = make_fs()
    for _, dnames, fnames in fs.walk(""):
        dirs.extend(dnames)
        files.extend(fnames)

    # currently we don't walk through submodules
    assert not dirs
    assert set(files) == {".gitmodules", "submodule"}


def test_ls(tmp_dir: TmpDir, scm: Git, make_fs):
    files = tmp_dir.gen(
        {
            "foo": "foo",
            "тест": "проверка",
            "data": {"lorem": "ipsum", "subdir": {"sub": "sub"}},
        }
    )
    scm.add_commit(files, message="add")
    fs = make_fs()

    assert fs.ls("/", detail=False) == ["/data", "/foo", "/тест"]
    assert fs.ls("/") == [
        {
            "mode": 16384,
            "name": "/data",
            "sha": "f5d6ac1955c85410b71bb6e35e4c57c54e2ad524",
            "size": 66,
            "type": "directory",
        },
        {
            "mode": 33188,
            "name": "/foo",
            "sha": "19102815663d23f8b75a47e7a01965dcdc96468c",
            "size": 3,
            "type": "file",
        },
        {
            "mode": 33188,
            "name": "/тест",
            "sha": "eeeba1738f4c12844163b89112070c6e57eb764e",
            "size": 16,
            "type": "file",
        },
    ]



from contextlib import AbstractContextManager
from types import TracebackType
from typing import TYPE_CHECKING, Callable, Optional, Union

from pygit2 import RemoteCallbacks as _RemoteCallbacks

from scmrepo.git.backend.base import SyncStatus
from scmrepo.git.credentials import Credential, CredentialNotFoundError
from scmrepo.progress import GitProgressReporter

if TYPE_CHECKING:
    from pygit2 import Oid
    from pygit2.credentials import Keypair, Username, UserPass
    from pygit2.enums import CredentialType

    from scmrepo.progress import GitProgressEvent


_Pygit2Credential = Union["Keypair", "Username", "UserPass"]


class RemoteCallbacks(_RemoteCallbacks, AbstractContextManager):
    def __init__(
        self,
        *args,
        progress: Optional[Callable[["GitProgressEvent"], None]] = None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.progress = GitProgressReporter(progress) if progress else None
        self._store_credentials: Optional[Credential] = None
        self._tried_credentials = False
        self.result: dict[str, SyncStatus] = {}

    def __exit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ):
        if exc_type is None:
            self._approve_credentials()

    def sideband_progress(self, string: str):
        if self.progress is not None:
            self.progress(string)

    def credentials(
        self,
        url: str,
        username_from_url: Optional[str],
        allowed_types: "CredentialType",
    ) -> "_Pygit2Credential":
        from pygit2 import GitError, Passthrough
        from pygit2.credentials import UserPass
        from pygit2.enums import CredentialType

        if self._tried_credentials:
            raise GitError(f"authentication failed for '{url}'")
        self._tried_credentials = True

        if allowed_types & CredentialType.USERPASS_PLAINTEXT:
            try:
                if self._store_credentials:
                    creds = self._store_credentials
                else:
                    creds = Credential(username=username_from_url, url=url).fill()
                    self._store_credentials = creds
                assert creds.username is not None
                assert creds.password is not None
                return UserPass(creds.username, creds.password)
            except CredentialNotFoundError:
                pass
        raise Passthrough

    def _approve_credentials(self):
        if self._store_credentials:
            self._store_credentials.approve()

    def update_tips(self, refname: str, old: "Oid", new: "Oid"):
        if old == new:
            self.result[refname] = SyncStatus.UP_TO_DATE
        else:
            self.result[refname] = SyncStatus.SUCCESS



"""asyncssh SSH vendor for Dulwich."""

import asyncio
import os
from collections.abc import Coroutine, Iterator
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Optional,
    cast,
)

from asyncssh import SSHClient
from dulwich.client import SSHVendor

from scmrepo.asyn import BaseAsyncObject, sync_wrapper
from scmrepo.exceptions import AuthError

if TYPE_CHECKING:
    from collections.abc import Sequence
    from pathlib import Path

    from asyncssh.auth import KbdIntPrompts, KbdIntResponse
    from asyncssh.config import ConfigPaths, FilePath
    from asyncssh.connection import SSHClientConnection
    from asyncssh.misc import MaybeAwait
    from asyncssh.process import SSHClientProcess
    from asyncssh.public_key import KeyPairListArg, SSHKey
    from asyncssh.stream import SSHReader


async def _read_all(read: Callable[[int], Coroutine], n: Optional[int] = None) -> bytes:
    if n is None:
        return await read(-1)
    result = []
    while n > 0:
        data = await read(n)
        result.append(data)
        n -= len(data)
    return b"".join(result)


async def _getpass(*args, **kwargs) -> str:
    from getpass import getpass

    return await asyncio.to_thread(getpass, *args, **kwargs)


class _StderrWrapper:
    def __init__(self, stderr: "SSHReader", loop: asyncio.AbstractEventLoop) -> None:
        self.stderr = stderr
        self.loop = loop

    async def _readlines(self) -> list[bytes]:
        lines = []
        while True:
            line = await self.stderr.readline()
            if not line:
                break
            lines.append(line)
        return lines

    async def _read(self, n: Optional[int] = None) -> bytes:
        if self.stderr.at_eof():
            return b""
        return await _read_all(self.stderr.read, n)

    read = sync_wrapper(_read)
    readlines = sync_wrapper(_readlines)


class AsyncSSHWrapper(BaseAsyncObject):
    def __init__(
        self, conn: "SSHClientConnection", proc: "SSHClientProcess", **kwargs
    ) -> None:
        super().__init__(**kwargs)
        self.conn: SSHClientConnection = conn
        self.proc: SSHClientProcess = proc
        self.stderr = _StderrWrapper(proc.stderr, self.loop)

    def can_read(self) -> bool:
        # pylint:disable=protected-access
        return self.proc.stdout._session._recv_buf_len > 0

    async def _read(self, n: Optional[int] = None) -> bytes:
        if self.proc.stdout.at_eof():
            return b""
        return await _read_all(self.proc.stdout.read, n)

    read = sync_wrapper(_read)

    def read_stderr(self, n: Optional[int] = None) -> bytes:
        return self.stderr.read(n)

    async def _write(self, data: bytes) -> None:
        self.proc.stdin.write(data)
        await self.proc.stdin.drain()

    write = sync_wrapper(_write)

    async def _close(self) -> None:
        self.conn.close()
        await self.conn.wait_closed()

    close = sync_wrapper(_close)


class InteractiveSSHClient(SSHClient):
    _conn: Optional["SSHClientConnection"] = None
    _keys_to_try: Optional[list["FilePath"]] = None
    _passphrases: dict[str, str]

    def __init__(self, *args, **kwargs) -> None:
        super(*args, **kwargs)
        self._passphrases: dict[str, str] = {}

    def connection_made(self, conn: "SSHClientConnection") -> None:
        self._conn = conn
        self._keys_to_try = None

    def connection_lost(self, exc: Optional[Exception]) -> None:
        self._conn = None

    async def public_key_auth_requested(  # noqa: C901
        self,
    ) -> Optional["KeyPairListArg"]:
        from asyncssh.public_key import (
            _DEFAULT_KEY_FILES,
            KeyImportError,
            SSHLocalKeyPair,
            read_private_key,
            read_public_key,
        )

        if os.environ.get("GIT_TERMINAL_PROMPT") == "0":
            return None

        assert self._conn is not None
        if self._keys_to_try is None:
            self._keys_to_try = []
            options = self._conn._options  # pylint: disable=protected-access
            config = options.config
            client_keys = cast("Sequence[FilePath]", config.get("IdentityFile", ()))
            if not client_keys:
                client_keys = [
                    os.path.expanduser(os.path.join("~", ".ssh", path))
                    for path, cond in _DEFAULT_KEY_FILES
                    if cond
                ]
            for key_to_load in client_keys:
                try:
                    read_private_key(key_to_load, passphrase=options.passphrase)
                except KeyImportError as exc:
                    if str(exc).startswith("Passphrase"):
                        self._keys_to_try.append(key_to_load)
                except OSError:
                    pass

        while self._keys_to_try:
            key_to_load = self._keys_to_try.pop()
            pubkey_to_load = str(key_to_load) + ".pub"
            try:
                key = await self._read_private_key_interactive(key_to_load)
            except KeyImportError:
                continue
            try:
                pubkey = read_public_key(pubkey_to_load)
            except (OSError, KeyImportError):
                pubkey = None
            return SSHLocalKeyPair(key, pubkey, cert=None, enc_key=None)
        return None

    async def _read_private_key_interactive(self, path: "FilePath") -> "SSHKey":
        from asyncssh.public_key import (
            KeyEncryptionError,
            KeyImportError,
            read_private_key,
        )

        path = str(path)
        passphrase = self._passphrases.get(path)
        if passphrase:
            return read_private_key(path, passphrase=passphrase)

        for _ in range(3):
            passphrase = await _getpass(f"Enter passphrase for key {path!r}: ")
            if passphrase:
                try:
                    key = read_private_key(path, passphrase=passphrase)
                    self._passphrases[path] = passphrase
                    return key
                except (KeyImportError, KeyEncryptionError):
                    pass
        raise KeyImportError("Incorrect passphrase")

    def kbdint_auth_requested(self) -> "MaybeAwait[Optional[str]]":
        return ""

    async def kbdint_challenge_received(  # pylint: disable=invalid-overridden-method
        self,
        name: str,
        instructions: str,
        lang: str,
        prompts: "KbdIntPrompts",
    ) -> Optional["KbdIntResponse"]:
        if os.environ.get("GIT_TERMINAL_PROMPT") == "0":
            return None

        if instructions:
            pass

        response: list[str] = []
        for prompt, _echo in prompts:
            p = await _getpass(f"({name}) {prompt}" if name else prompt)
            response.append(p.rstrip())
        return response

    async def password_auth_requested(self) -> str:
        return await _getpass()


class AsyncSSHVendor(BaseAsyncObject, SSHVendor):
    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

    async def _run_command(
        self,
        host: str,
        command: list[str],
        username: Optional[str] = None,
        port: Optional[int] = None,
        password: Optional[str] = None,
        key_filename: Optional[str] = None,
        **kwargs: object,
    ) -> AsyncSSHWrapper:
        """Connect to an SSH server.

        Run a command remotely and return a file-like object for interaction
        with the remote command.

        Args:
          host: Host name
          command: Command to run (as argv array)
          username: Optional ame of user to log in as
          port: Optional SSH port to use
          password: Optional ssh password for login or private key
          key_filename: Optional path to private keyfile
        """
        import asyncssh

        try:
            conn = await asyncssh.connect(
                host,
                port=port if port is not None else (),
                username=username if username is not None else (),
                password=password,
                client_keys=[key_filename] if key_filename else (),
                ignore_encrypted=not key_filename,
                known_hosts=None,
                encoding=None,
                client_factory=InteractiveSSHClient,
            )
            proc: SSHClientProcess[Any] = await conn.create_process(
                command, encoding=None
            )
        except asyncssh.misc.PermissionDenied as exc:
            raise AuthError(f"{username}@{host}:{port or 22}") from exc
        return AsyncSSHWrapper(conn, proc)

    run_command = sync_wrapper(_run_command)


def get_unsupported_opts(config_paths: "ConfigPaths") -> Iterator[str]:
    from pathlib import Path, PurePath

    if config_paths:
        if isinstance(config_paths, (str, PurePath)):
            paths: Sequence[FilePath] = [config_paths]
        else:
            paths = config_paths

        for path in paths:
            try:
                yield from _parse_unsupported(Path(path))
            except FileNotFoundError:
                continue


def _parse_unsupported(path: "Path") -> Iterator[str]:
    import locale
    import shlex

    from asyncssh.config import SSHClientConfig

    handlers = SSHClientConfig._handlers  # pylint: disable=protected-access
    with open(path, encoding=locale.getpreferredencoding()) as fobj:
        for line in fobj:
            line = line.strip()
            if not line or line[0] == "#":
                continue

            try:
                args = shlex.split(line)
            except ValueError:
                continue

            option = args.pop(0)
            if option.endswith("="):
                option = option[:-1]
            elif "=" in option:
                option, _ = option.split("=", 1)
            loption = option.lower()
            if loption not in handlers:
                yield loption



"""Manages source control systems (e.g. Git) in DVC."""

from contextlib import AbstractContextManager


class Base(AbstractContextManager):
    """Base class for source control management driver implementations."""

    def __init__(self, root_dir=None):
        import os

        self._root_dir = os.path.realpath(root_dir or os.curdir)

    @property
    def root_dir(self) -> str:
        return self._root_dir

    def __repr__(self):
        return f"{type(self).__name__}: '{self.dir}'"

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    @property
    def dir(self):
        """Path to a directory with SCM specific information."""
        return None

    @staticmethod
    def is_repo(root_dir):  # pylint: disable=unused-argument
        """Returns whether or not root_dir is a valid SCM repository."""
        return True

    @staticmethod
    def is_submodule(root_dir):  # pylint: disable=unused-argument
        """Returns whether or not root_dir is a valid SCM repository
        submodule.
        """
        return True

    def is_ignored(self, path):  # pylint: disable=unused-argument
        """Returns whether or not path is ignored by SCM."""
        return False

    def ignore(self, path):  # pylint: disable=unused-argument
        """Makes SCM ignore a specified path."""

    def ignore_remove(self, path):  # pylint: disable=unused-argument
        """Makes SCM stop ignoring a specified path."""

    @property
    def ignore_file(self):
        """Filename for a file that contains ignored paths for this SCM."""

    def add(self, paths):
        """Makes SCM track every path from a specified list of paths."""

    def commit(self, msg):
        """Makes SCM create a commit."""

    def checkout(self, branch, create_new=False):
        """Makes SCM checkout a branch."""

    def branch(self, branch):
        """Makes SCM create a branch with a specified name."""

    def tag(self, tag):
        """Makes SCM create a tag with a specified name."""

    def untracked_files(self):
        """Returns a list of untracked files."""
        return []

    def is_tracked(self, path):  # pylint: disable=unused-argument
        """Returns whether or not a specified path is tracked."""
        return False

    def is_dirty(self):
        """Return whether the SCM contains uncommitted changes."""
        return False

    def active_branch(self):
        """Returns current branch in the repo."""
        return ""

    def list_branches(self):
        """Returns a list of available branches in the repo."""
        return []

    def list_tags(self):
        """Returns a list of available tags in the repo."""
        return []

    def list_all_commits(self):
        """Returns a list of commits in the repo."""
        return []

    def belongs_to_scm(self, path):
        """Return boolean whether file belongs to scm"""

    def close(self):
        """Method to close the files"""

    def _reset(self) -> None:
        pass

    def list_submodules(self) -> list[str]:
        """Returns a list of submodules in the repo."""
        return []



# pylint: disable=redefined-outer-name
import io
from collections import defaultdict
from collections.abc import Sequence
from http import HTTPStatus
from time import time
from typing import Callable

import pytest
from aiohttp import ClientResponseError
from aioresponses import CallbackResult, aioresponses
from pytest_mock import MockerFixture
from pytest_test_utils import TempDirFactory, TmpDir
from yarl import URL

from scmrepo.git import Git
from scmrepo.git.lfs import LFSClient, LFSStorage, Pointer, smudge

FOO_OID = "2c26b46b68ffc68ff99b453c1d30413413422d706483bfa0f98a5e886266e7ae"
FOO_SIZE = 3
FOO_POINTER = (
    f"version https://git-lfs.github.com/spec/v1\n"
    f"oid sha256:{FOO_OID}\n"
    f"size {FOO_SIZE}\n"
).encode()


@pytest.fixture
def storage(tmp_dir_factory: TempDirFactory) -> LFSStorage:
    storage_path = tmp_dir_factory.mktemp("lfs")
    return LFSStorage(storage_path)


@pytest.fixture
def lfs(tmp_dir: TmpDir, scm: Git) -> None:
    tmp_dir.gen(".gitattributes", "*.lfs filter=lfs diff=lfs merge=lfs -text")
    scm.add([".gitattributes"])
    scm.commit("init lfs attributes")


@pytest.fixture
def lfs_objects(tmp_dir: TmpDir) -> TmpDir:
    objects = tmp_dir / ".git" / "lfs" / "objects"
    objects.mkdir(parents=True)
    return objects


def test_pointer_build(tmp_dir: TmpDir):
    tmp_dir.gen("foo", "foo")
    with open(tmp_dir / "foo", "rb") as fobj:
        pointer = Pointer.build(fobj)

    assert pointer.dump() == FOO_POINTER.decode("utf-8")


def test_pointer_load(tmp_dir: TmpDir):
    tmp_dir.gen("foo.lfs", FOO_POINTER)
    with open(tmp_dir / "foo.lfs", "rb") as fobj:
        pointer = Pointer.load(fobj)
    assert pointer.oid == FOO_OID
    assert pointer.size == 3


def test_smudge(tmp_dir: TmpDir, storage: LFSStorage, mocker: MockerFixture):
    tmp_dir.gen("foo.lfs", FOO_POINTER)
    with open(tmp_dir / "foo.lfs", "rb") as fobj:
        assert smudge(storage, fobj).read() == FOO_POINTER

    mocker.patch.object(storage, "open", return_value=io.BytesIO(b"foo"))
    with open(tmp_dir / "foo.lfs", "rb") as fobj:
        assert smudge(storage, fobj).read() == b"foo"


@pytest.mark.usefixtures("lfs")
def test_lfs(tmp_dir: TmpDir, scm: Git, lfs_objects: TmpDir):
    # NOTE: scmrepo does not currently support LFS clean (writes), this writes
    # the pointer to the git odb (simulating an actual LFS clean)
    tmp_dir.gen("foo.lfs", FOO_POINTER)
    scm.add(["foo.lfs"])
    scm.commit("add foo")
    lfs_objects.gen({FOO_OID[0:2]: {FOO_OID[2:4]: {FOO_OID: "foo"}}})

    fs = scm.get_fs("HEAD")
    with fs.open("foo.lfs", "rb", raw=True) as fobj:
        assert fobj.read() == FOO_POINTER
    with fs.open("foo.lfs", "rb", raw=False) as fobj:
        assert fobj.read() == b"foo"


class CallbackResultRecorder:
    def __init__(self) -> None:
        self._results: dict[str, list[CallbackResult]] = defaultdict(list)

    def record(self, result: CallbackResult) -> Callable[..., CallbackResult]:
        def _callback(url: URL, **_) -> CallbackResult:
            self._results[str(url)].append(result)
            return result

        return _callback

    def __getitem__(self, url: str) -> Sequence[CallbackResult]:
        return self._results[url]


class LFSServerMock:
    def __init__(
        self,
        mocker: aioresponses,
        recorder: CallbackResultRecorder,
        batch_url: str,
        objects_url: str,
    ) -> None:
        self._mocker = mocker
        self._recorder = recorder
        self.batch_url = batch_url
        self.objects_url = objects_url

    def get_object_url(self, oid: str) -> str:
        return f"{self.objects_url}/{oid}"

    def mock_batch_200(self, pointer: Pointer) -> None:
        self._mocker.post(
            self.batch_url,
            callback=self._recorder.record(
                CallbackResult(
                    status=HTTPStatus.OK,
                    headers={"Content-Type": "application/vnd.git-lfs+json"},
                    payload={
                        "transfer": "basic",
                        "objects": [
                            {
                                "oid": pointer.oid,
                                "size": pointer.size,
                                "authenticated": True,
                                "actions": {
                                    "download": {
                                        "href": self.get_object_url(pointer.oid),
                                    }
                                },
                            }
                        ],
                        "hash_algo": "sha256",
                    },
                )
            ),
        )

    def mock_batch_429(
        self, header: str, value: Callable[[], str], *, repeat: bool = False
    ) -> None:
        self._mocker.post(
            self.batch_url,
            callback=self._recorder.record(
                CallbackResult(
                    status=HTTPStatus.TOO_MANY_REQUESTS,
                    headers={header: value()},
                    reason="Too many requests",
                )
            ),
            repeat=repeat,
        )

    def mock_object_200(self, oid: str) -> None:
        self._mocker.get(
            self.get_object_url(oid),
            callback=self._recorder.record(
                CallbackResult(
                    status=HTTPStatus.OK,
                    body=f"object {oid} data",
                )
            ),
        )

    def mock_object_429(
        self,
        oid: str,
        header: str,
        value: Callable[[], str],
        *,
        repeat: bool = False,
    ) -> None:
        self._mocker.get(
            self.get_object_url(oid),
            callback=self._recorder.record(
                CallbackResult(
                    status=HTTPStatus.TOO_MANY_REQUESTS,
                    headers={header: value()},
                    reason="Too many requests",
                )
            ),
            repeat=repeat,
        )


@pytest.mark.parametrize(
    "rate_limit_header, rate_limit_value",
    [
        ("Retry-After", lambda: "1"),
        ("RateLimit-Reset", lambda: f"{int(time()) + 1}"),
        ("X-RateLimit-Reset", lambda: f"{int(time()) + 1}"),
    ],
)
@pytest.mark.filterwarnings("ignore:enable_cleanup_closed ignored.*:DeprecationWarning")
def test_rate_limit_retry(
    storage: LFSStorage, rate_limit_header: str, rate_limit_value: Callable[[], str]
):
    client = LFSClient.from_git_url("http://git.example.com/namespace/project.git")
    recorder = CallbackResultRecorder()

    with aioresponses() as m:
        lfs_server = LFSServerMock(
            m, recorder, f"{client.url}/objects/batch", "http://git-lfs.example.com"
        )
        lfs_server.mock_batch_429(rate_limit_header, rate_limit_value)
        lfs_server.mock_batch_200(Pointer(FOO_OID, FOO_SIZE))
        lfs_server.mock_object_429(FOO_OID, rate_limit_header, rate_limit_value)
        lfs_server.mock_object_200(FOO_OID)

        client.download(storage, [Pointer(oid=FOO_OID, size=FOO_SIZE)])

        results = recorder[lfs_server.batch_url]
        assert [r.status for r in results] == [429, 200]

        results = recorder[lfs_server.get_object_url(FOO_OID)]
        assert [r.status for r in results] == [429, 200]


@pytest.mark.parametrize(
    "rate_limit_header, rate_limit_value",
    [
        ("Retry-After", lambda: "1"),
        ("RateLimit-Reset", lambda: f"{int(time()) + 1}"),
        ("X-RateLimit-Reset", lambda: f"{int(time()) + 1}"),
    ],
)
@pytest.mark.filterwarnings("ignore:enable_cleanup_closed ignored.*:DeprecationWarning")
def test_rate_limit_max_retries_batch(
    storage: LFSStorage, rate_limit_header: str, rate_limit_value: Callable[[], str]
):
    client = LFSClient.from_git_url("http://git.example.com/namespace/project.git")
    recorder = CallbackResultRecorder()

    with aioresponses() as m:
        lfs_server = LFSServerMock(
            m, recorder, f"{client.url}/objects/batch", "http://git-lfs.example.com"
        )
        lfs_server.mock_batch_429(rate_limit_header, rate_limit_value, repeat=True)

        with pytest.raises(ClientResponseError, match="Too many requests"):
            client.download(storage, [Pointer(oid=FOO_OID, size=FOO_SIZE)])

        results = recorder[lfs_server.batch_url]
        assert [r.status for r in results] == [429] * 5


@pytest.mark.parametrize(
    "rate_limit_header, rate_limit_value",
    [
        ("Retry-After", lambda: "1"),
        ("RateLimit-Reset", lambda: f"{int(time()) + 1}"),
        ("X-RateLimit-Reset", lambda: f"{int(time()) + 1}"),
    ],
)
@pytest.mark.filterwarnings("ignore:enable_cleanup_closed ignored.*:DeprecationWarning")
def test_rate_limit_max_retries_objects(
    storage: LFSStorage, rate_limit_header: str, rate_limit_value: Callable[[], str]
):
    client = LFSClient.from_git_url("http://git.example.com/namespace/project.git")
    recorder = CallbackResultRecorder()

    with aioresponses() as m:
        lfs_server = LFSServerMock(
            m, recorder, f"{client.url}/objects/batch", "http://git-lfs.example.com"
        )
        lfs_server.mock_batch_200(Pointer(FOO_OID, FOO_SIZE))
        lfs_server.mock_object_429(
            FOO_OID, rate_limit_header, rate_limit_value, repeat=True
        )

        with pytest.raises(ClientResponseError, match="Too many requests"):
            client.download(storage, [Pointer(oid=FOO_OID, size=FOO_SIZE)])

        results = recorder[lfs_server.batch_url]
        assert [r.status for r in results] == [200]

        results = recorder[lfs_server.get_object_url(FOO_OID)]
        assert [r.status for r in results] == [429] * 5



from typing import NamedTuple, Optional, Union

from funcy import compose


def code2desc(op_code):
    from git import RootUpdateProgress as OP  # noqa: N814, TID251

    ops = {
        OP.COUNTING: "Counting",
        OP.COMPRESSING: "Compressing",
        OP.WRITING: "Writing",
        OP.RECEIVING: "Receiving",
        OP.RESOLVING: "Resolving",
        OP.FINDING_SOURCES: "Finding sources",
        OP.CHECKING_OUT: "Checking out",
        OP.CLONE: "Cloning",
        OP.FETCH: "Fetching",
        OP.UPDWKTREE: "Updating working tree",
        OP.REMOVE: "Removing",
        OP.PATHCHANGE: "Changing path",
        OP.URLCHANGE: "Changing URL",
        OP.BRANCHCHANGE: "Changing branch",
    }
    return ops.get(op_code & OP.OP_MASK, "")


class GitProgressEvent(NamedTuple):
    phase: str = ""
    completed: Optional[int] = None
    total: Optional[int] = None
    message: str = ""

    @classmethod
    def parsed_from_gitpython(
        cls,
        op_code,
        cur_count,
        max_count=None,
        message="",  # pylint: disable=redefined-outer-name
    ):
        return cls(code2desc(op_code), cur_count, max_count, message)


class GitProgressReporter:
    def __init__(self, fn) -> None:
        try:
            from git.util import CallableRemoteProgress  # noqa: TID251
        except ImportError:
            self._reporter = None
        else:
            self._reporter = CallableRemoteProgress(self.wrap_fn(fn))

    def __call__(self, msg: Union[str, bytes]) -> None:
        if self._reporter is not None:
            self._reporter._parse_progress_line(
                msg.decode("utf-8", errors="replace").strip()
                if isinstance(msg, bytes)
                else msg
            )

    @staticmethod
    def wrap_fn(fn):
        return compose(fn, GitProgressEvent.parsed_from_gitpython)


if __name__ == "__main__":
    message = """
Cloning into 'dvcyaml-schema'...
remote: Enumerating objects: 76, done.
remote: Counting objects: 100% (76/76), done.
remote: Compressing objects: 100% (56/56), done.
remote: Total 76 (delta 30), reused 38 (delta 9), pack-reused 0
Receiving objects: 100% (76/76), 24.83 KiB | 89.00 KiB/s, done.
Resolving deltas: 100% (30/30), done."""

    reporter = GitProgressReporter(print)
    for line in message.splitlines():
        reporter(line)



import pytest

from scmrepo.urls import is_scp_style_url


@pytest.mark.parametrize(
    "url",
    [
        "git@github.com:treeverse/scmrepo.git",
        "github.com:treeverse/scmrepo.git",
        "user@github.com:treeverse/scmrepo.git",
    ],
)
def test_scp_url(url: str):
    assert is_scp_style_url(url)


@pytest.mark.parametrize(
    "url",
    [
        r"C:\foo\bar",
        "C:/foo/bar",
        "/home/user/treeverse/scmrepo/git",
        "~/treeverse/scmrepo/git",
        "ssh://login@server.com:12345/repository.git",
        "https://user:password@github.com/treeverse/scmrepo.git",
        "https://github.com/treeverse/scmrepo.git",
    ],
)
def test_scp_url_invalid(url: str):
    assert not is_scp_style_url(url)



import io
import logging
from typing import TYPE_CHECKING, BinaryIO, Optional

from .pointer import HEADERS, Pointer

if TYPE_CHECKING:
    from .storage import LFSStorage

logger = logging.getLogger(__name__)


def smudge(
    storage: "LFSStorage",
    fobj: BinaryIO,
    url: Optional[str] = None,
    batch_size: Optional[int] = None,
) -> BinaryIO:
    """Wrap the specified binary IO stream and run LFS smudge if necessary."""
    reader = io.BufferedReader(fobj)  # type: ignore[type-var]
    data = reader.peek(100)
    if any(data.startswith(header) for header in HEADERS):
        # read the pointer data into memory since the raw stream is unseekable
        # and we may need to return it in fallback case
        data = reader.read()
        lfs_obj: Optional[BinaryIO] = None
        try:
            pointer = Pointer.load(io.BytesIO(data))
            lfs_obj = storage.open(pointer, mode="rb", fetch_url=url)
        except (ValueError, OSError):
            logger.warning("Could not open LFS object, falling back to raw pointer")
        if lfs_obj:
            return lfs_obj
        return io.BytesIO(data)
    return reader


if __name__ == "__main__":
    # Minimal `git lfs smudge` CLI implementation
    import sys

    from scmrepo.git import Git

    if sys.stdin.isatty():
        sys.exit(
            "Cannot read from STDIN: "
            "This command should be run by the Git 'smudge' filter"
        )
    scm = Git()
    try:
        with smudge(scm.lfs_storage, sys.stdin.buffer) as fobj_:
            sys.stdout.buffer.write(fobj_.read())
    finally:
        scm.close()



from unittest.mock import MagicMock

from pytest_test_utils import TmpDir
from pytest_test_utils.matchers import Matcher

from scmrepo.git import Git
from scmrepo.progress import GitProgressEvent


def test_clone(tmp_dir: TmpDir, matcher: type[Matcher]):
    progress = MagicMock()
    url = "https://github.com/treeverse/dvcyaml-schema"
    rev = "cf279597596b54c5b0ce089eb4bda41ebbbb5db4"

    repo = Git.clone(url, "dir", rev=rev, progress=progress)
    assert repo.get_rev() == rev

    progress.assert_called_with(matcher.instance_of(GitProgressEvent))
    assert (tmp_dir / "dir").exists()


def test_clone_shallow(tmp_dir: TmpDir):
    url = "https://github.com/treeverse/dvcyaml-schema"
    shallow_branch = "master"

    repo = Git.clone(url, "dir", shallow_branch=shallow_branch)
    shallow_file = tmp_dir / "dir" / ".git" / "shallow"
    assert shallow_file.exists()
    assert repo.get_rev() in shallow_file.read_text().splitlines()



class SCMError(Exception):
    """Base class for source control management errors."""


class GitHookAlreadyExists(SCMError):
    def __init__(self, name: str) -> None:
        self.name = name
        super().__init__(f"Hook '{name}' already exists")


class FileNotInRepoError(SCMError):
    """Thrown when trying to find .gitignore for a file that is not in a scm
    repository.
    """


class MergeConflictError(SCMError):
    pass


class InvalidRemote(SCMError):
    def __init__(self, url: str) -> None:
        self.url = url
        super().__init__(f"'{url}' is not a valid Git remote or URL")


class AuthError(SCMError):
    def __init__(self, url: str) -> None:
        self.url = url
        super().__init__(f"Authentication failed for: '{url}'")


class CloneError(SCMError):
    def __init__(self, url: str, path: str) -> None:
        self.url = url
        self.path = path
        super().__init__(f"Failed to clone repo '{url}' to '{path}'")


class RevError(SCMError):
    pass


class UnsupportedIndexFormat(SCMError):
    pass
